#!/usr/bin/env python
# encoding: utf-8

r"""
Shallow water flow
==================

Solve the one-dimensional shallow water equations:

.. math::
    h_t + (hu)_x & = 0 \\
    (hu)_t + (hu^2 + \frac{1}{2}gh^2)_x & = 0.

Here h is the depth, u is the velocity, and g is the gravitational constant.
The default initial condition used here models a dam break.
"""

import numpy as np
from clawpack import riemann
from clawpack.riemann.shallow_roe_with_efix_1D_constants import depth, momentum, num_eqn

def setup(use_petsc=False,kernel_language='Fortran',outdir='./_output',solver_type='classic',
          riemann_solver='roe', disable_output=False):

    if use_petsc:
        import clawpack.petclaw as pyclaw
    else:
        from clawpack import pyclaw

    if kernel_language == 'Python':
        if riemann_solver.lower() == 'roe':
            raise Exception('Python Roe solver not implemented.')
        elif riemann_solver.lower() == 'hlle':
            rs = riemann.shallow_1D_py.shallow_hll_1D
    elif kernel_language == 'Fortran':
        if riemann_solver.lower() == 'roe':
            rs = riemann.shallow_roe_with_efix_1D
        elif riemann_solver.lower() == 'hlle':
            rs = riemann.shallow_hlle_1D
 
    if solver_type == 'classic':
        solver = pyclaw.ClawSolver1D(rs)
        solver.limiters = pyclaw.limiters.tvd.vanleer
    elif solver_type == 'sharpclaw':
        solver = pyclaw.SharpClawSolver1D(rs)

    solver.kernel_language = kernel_language

    solver.bc_lower[0] = pyclaw.BC.extrap
    solver.bc_upper[0] = pyclaw.BC.extrap

    xlower = -5.0
    xupper = 5.0
    mx = 500
    x = pyclaw.Dimension(xlower,xupper,mx,name='x')
    domain = pyclaw.Domain(x)
    state = pyclaw.State(domain,num_eqn)

    # Gravitational constant
    state.problem_data['grav'] = 1.0
    state.problem_data['dry_tolerance'] = 1e-3
    state.problem_data['sea_level'] = 0.0
    
    xc = state.grid.x.centers

    IC='dam-break'
    x0=0.

    if IC=='dam-break':
        hl = 3.
        ul = 0.
        hr = 1.
        ur = 0.
        state.q[depth,:] = hl * (xc <= x0) + hr * (xc > x0)
        state.q[momentum,:] = hl*ul * (xc <= x0) + hr*ur * (xc > x0)
    elif IC=='2-shock':
        hl = 1.
        ul = 1.
        hr = 1.
        ur = -1.
        state.q[depth,:] = hl * (xc <= x0) + hr * (xc > x0)
        state.q[momentum,:] = hl*ul * (xc <= x0) + hr*ur * (xc > x0)
    elif IC=='perturbation':
        eps=0.1
        state.q[depth,:] = 1.0 + eps*np.exp(-(xc-x0)**2/0.5)
        state.q[momentum,:] = 0.

    claw = pyclaw.Controller()
    claw.keep_copy = True
    if disable_output:
        claw.output_format = None
    claw.tfinal = 2.0
    claw.solution = pyclaw.Solution(state,domain)
    claw.solver = solver
    claw.outdir = outdir
    claw.setplot = setplot

    return claw


#--------------------------
def setplot(plotdata):
#--------------------------
    """ 
    Specify what is to be plotted at each frame.
    Input:  plotdata, an instance of visclaw.data.ClawPlotData.
    Output: a modified version of plotdata.
    """ 
    plotdata.clearfigures()  # clear any old figures,axes,items data

    # Figure for depth
    plotfigure = plotdata.new_plotfigure(name='Water height', figno=0)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = [-5.0,5.0]
    plotaxes.title = 'Water height'
    plotaxes.axescmd = 'subplot(211)'

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='1d')
    plotitem.plot_var = depth
    plotitem.plotstyle = '-'
    plotitem.color = 'b'
    plotitem.kwargs = {'linewidth':3}

    # Figure for momentum[1]
    #plotfigure = plotdata.new_plotfigure(name='Momentum', figno=1)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.axescmd = 'subplot(212)'
    plotaxes.xlimits = [-5.0,5.0]
    plotaxes.title = 'Momentum'

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='1d')
    plotitem.plot_var = momentum
    plotitem.plotstyle = '-'
    plotitem.color = 'b'
    plotitem.kwargs = {'linewidth':3}
    
    return plotdata


if __name__=="__main__":
    from clawpack.pyclaw.util import run_app_from_main
    output = run_app_from_main(setup,setplot)



#!/usr/bin/env python
# coding: utf-8

# # process_fgmax
# 
# Read in fgmax results and produce plots.

# In[ ]:


from pylab import *


# In[ ]:


import os,sys
import glob
from scipy.interpolate import RegularGridInterpolator
import matplotlib as mpl
from matplotlib import colors

from clawpack.geoclaw import topotools, dtopotools
from clawpack.visclaw import colormaps
from clawpack.visclaw.plottools import pcolorcells
from clawpack.geoclaw import fgmax_tools


# In[ ]:


save_figs = True
fgmax_plotdir = '_plots/fgmax_plots'


# In[ ]:


os.system('mkdir -p %s' % fgmax_plotdir)
def savefigp(fname):
    global save_figs
    if save_figs:
        fullname = '%s/%s' % (fgmax_plotdir, fname)
        savefig(fullname)
        print('Created ', fullname)
    else:
        print('save_figs = False')


# In[ ]:


outdir = '_output'
t_files = glob.glob(outdir + '/fort.t0*')
times = []
for f in t_files:
    lines = open(f,'r').readlines()
    for line in lines:
        if 'time' in line: 
            t = float(line.split()[0])
    times.append(t)
times.sort()
print('Output times found: ',times)
if len(times) > 0:
    t_hours = times[-1] / 3600.
    print('\nfgmax results are presumably from final time: %.1f seconds = %.2f hours'          % (times[-1], t_hours))
else:
    t_hours = nan


# In[ ]:


# Read fgmax data:
fgno = 1
fg = fgmax_tools.FGmaxGrid()
fg.read_fgmax_grids_data(fgno)

fg.read_output(outdir=outdir)


# In[ ]:


zmin = -60.
zmax = 20.
land_cmap = colormaps.make_colormap({ 0.0:[0.1,0.4,0.0],
                                     0.25:[0.0,1.0,0.0],
                                      0.5:[0.8,1.0,0.5],
                                      1.0:[0.8,0.5,0.2]})

sea_cmap = colormaps.make_colormap({ 0.0:[0,0,1], 1.:[.8,.8,1]})

cmap, norm = colormaps.add_colormaps((land_cmap, sea_cmap),
                                     data_limits=(zmin,zmax),
                                     data_break=0.)                                   



figure(figsize=(8,8))
pc = pcolorcells(fg.X, fg.Y, fg.B, cmap=cmap, norm=norm)  

cb = colorbar(pc,shrink=0.5,extend='both')
cb.set_label('meters')
cb.set_ticks(hstack((linspace(zmin,0,5), linspace(0,zmax,5))))

gca().set_aspect(1./cos(48*pi/180.))
ticklabel_format(useOffset=False)
xticks(rotation=20);
    
title('GeoClaw B topography on fg1 grid');


# In[ ]:


fg.B0 = fg.B  # no seafloor deformation in this problem
fg.h_onshore = ma.masked_where(fg.B0 < 0., fg.h)


# In[ ]:


bounds_depth = array([1e-6,0.5,1.0,1.5,2,2.5,3.0])


cmap_depth = colors.ListedColormap([[.7,.7,1],[.5,.5,1],[0,0,1],                 [1,.7,.7], [1,.4,.4], [1,0,0]])

# Set color for value exceeding top of range to purple:
cmap_depth.set_over(color=[1,0,1])

# Set color for land points without inundation to light green:
cmap_depth.set_under(color=[.7,1,.7])

norm_depth = colors.BoundaryNorm(bounds_depth, cmap_depth.N)
    

figure(figsize=(8,8))
pc = pcolorcells(fg.X, fg.Y, fg.h_onshore, cmap=cmap_depth, norm=norm_depth)
cb = colorbar(pc, extend='max', shrink=0.7)
cb.set_label('meters')
contour(fg.X, fg.Y, fg.B, [0], colors='g')

gca().set_aspect(1./cos(48*pi/180.))
ticklabel_format(useOffset=False)
xticks(rotation=20)
title('Maximum Onshore flow depth over %.2f hours\nfgmax grid %s' % (t_hours,fgno))
savefigp('fgmax%s_h_onshore.png' % str(fgno).zfill(4))


# In[ ]:


bounds_speed = np.array([1e-6,0.5,1.0,1.5,2,2.5,3,4.5,6])
cmap_speed = mpl.colors.ListedColormap([[.9,.9,1],[.6,.6,1],                 [.3,.3,1],[0,0,1], [1,.8,.8],                 [1,.6,.6], [1,.3,.3], [1,0,0]])


bounds_speed = np.array([1e-6,0.5,1.0,1.5,2,2.5,3,4.5])
cmap_speed = mpl.colors.ListedColormap([[.9,.9,1],[.6,.6,1],                 [.3,.3,1],[0,0,1], [1,.8,.8],                 [1,.6,.6], [1,0,0]])

# Set color for value exceeding top of range to purple:
cmap_speed.set_over(color=[1,0,1])

# Set color for land points without inundation to light green:
cmap_speed.set_under(color=[.7,1,.7])

norm_speed = colors.BoundaryNorm(bounds_speed, cmap_speed.N)

figure(figsize=(8,8))
pc = pcolorcells(fg.X, fg.Y, fg.s, cmap=cmap_speed, norm=norm_speed)
cb = colorbar(pc, extend='max', shrink=0.7)
cb.set_label('m/s')
contour(fg.X, fg.Y, fg.B0, [0], colors='g')

gca().set_aspect(1./cos(48*pi/180.))
ticklabel_format(useOffset=False)
xticks(rotation=20)
title('Maximum speed over %.2f hours\nfgmax grid %s' % (t_hours,fgno))
savefigp('fgmax%s_speed.png' % str(fgno).zfill(4))


# Save this so we can plot the topo below...

# In[ ]:


import copy
fg1 = copy.copy(fg)


# ## Read fgmax values specified on a Transect

# In[ ]:


# Read fgmax data:
fgno = 2
fg = fgmax_tools.FGmaxGrid()
fg.read_fgmax_grids_data(fgno)
fg.read_output(outdir=outdir)
xx = fg.X
yy = fg.Y


# In[ ]:


figure(figsize=(8,8))
pc = pcolorcells(fg1.X, fg1.Y, fg1.B, cmap=cmap, norm=norm)  

cb = colorbar(pc,shrink=0.5,extend='both')
cb.set_label('meters')
cb.set_ticks(hstack((linspace(zmin,0,5), linspace(0,zmax,5))))

gca().set_aspect(1./cos(48*pi/180.))
ticklabel_format(useOffset=False)
xticks(rotation=20);
    
plot(xx,yy,'r')
title('GeoClaw B topography values on fg1 grid\n with transect from fg2');


# In[ ]:


figure(figsize=(12,4))
fill_between(xx, fg.B, fg.B+fg.h, color=[.5,.5,1])
plot(xx,fg.B+fg.h,'b')
plot(xx,fg.B,'g')
plot(xx, ma.masked_where(fg.B>0, 0*xx), 'k')
grid(True)
ylim(-10,20);
title('Maximum elevation over %.2f hours\nfgmax grid %s' % (t_hours,fgno))
savefigp('fgmax%s_surface.png' % str(fgno).zfill(4));


# ## Read fgmax points as specified on a masked grid

# In[ ]:


fgno = 3
fg = fgmax_tools.FGmaxGrid()
fg.read_fgmax_grids_data(fgno)

fg.read_output(outdir=outdir)


# In[ ]:


fg.B0 = fg.B  # no seafloor deformation in this problem
fg.h_onshore = ma.masked_where(fg.B0 < 0., fg.h)


# In[ ]:


figure(figsize=(8,8))
pc = pcolorcells(fg.X, fg.Y, fg.B, cmap=cmap, norm=norm)
cb = colorbar(pc, extend='both', shrink=0.7)
cb.set_label('meters')
cb.set_ticks(hstack((linspace(zmin,0,5), linspace(0,zmax,5))))

gca().set_aspect(1./cos(48*pi/180.))
ticklabel_format(useOffset=False)
xticks(rotation=20)
title('GeoClaw B at points selected as fgmax grid\nfgmax grid %s' % fgno);


# In[ ]:


figure(figsize=(8,8))
pc = pcolorcells(fg.X, fg.Y, fg.h_onshore, cmap=cmap_depth, norm=norm_depth)
cb = colorbar(pc, extend='max', shrink=0.7)
cb.set_label('meters')
contour(fg.X, fg.Y, fg.B0, [0], colors='g')

gca().set_aspect(1./cos(48*pi/180.))
ticklabel_format(useOffset=False)
xticks(rotation=20)
title('Maximum Onshore flow depth over %.2f hours' % t_hours);
savefigp('fgmax%s_h_onshore.png' % str(fgno).zfill(4))


# In[ ]:


figure(figsize=(8,8))
pc = pcolorcells(fg.X, fg.Y, fg.s, cmap=cmap_speed, norm=norm_speed)
cb = colorbar(pc, extend='max', shrink=0.7)
cb.set_label('m/s')
contour(fg.X, fg.Y, fg.B0, [0], colors='g')

gca().set_aspect(1./cos(48*pi/180.))
ticklabel_format(useOffset=False)
xticks(rotation=20)
title('Maximum speed over %.2f hours\nfgmax grid %s' % (t_hours,fgno))
savefigp('fgmax%s_speed.png' % str(fgno).zfill(4))


# ### View fgmax points selected
# 
# This isn't generally needed, but if you want to inspect the file that specified fgmax points originally:

# In[ ]:


fg3input = topotools.Topography(path=fg.xy_fname, topo_type=3)
fg3input.X.shape

figure(figsize=(8,8))
pc = pcolorcells(fg3input.X, fg3input.Y, fg3input.Z)
cb = colorbar(pc, shrink=0.7)

gca().set_aspect(1./cos(48*pi/180.))
ticklabel_format(useOffset=False)
xticks(rotation=20);


# ## Read points with `point_style == 0`

# In[ ]:


# Read fgmax data:
fg = fgmax_tools.FGmaxGrid()
fg.read_fgmax_grids_data(4)

fg.read_output(outdir=outdir)
print('\n      x          y       max depth')
for j in range(fg.npts):
    print('%10.3f %10.3f %10.3f'  % (fg.X[j], fg.Y[j], fg.h[j]))


# In[ ]:


# Read fgmax data:
fg = fgmax_tools.FGmaxGrid()
fg.read_fgmax_grids_data(5)

fg.read_output(outdir=outdir)
print('\n      x          y       max speed')
for j in range(fg.npts):
    print('%10.3f %10.3f %10.3f'  % (fg.X[j], fg.Y[j], fg.s[j]))


# In[ ]:







#!/usr/bin/python
"""
iotools
=========
    Provides several useful functions for easy input and output from data files

    Contains:
        convertd2e
        datafile2array
        array2datafile


"""

from __future__ import absolute_import
import string
import re
import numpy


#====================================================================
def convertd2e (numberstring=" "):
    """
    def convertd2e (numberstring):

        takes a string and replaces all d's or D's with e's.
        usefule for reading data files with doubles output by 
        Fortran, which uses d, or D instead of e, for the exponent.
    """
    Dd=re.compile("[Dd]")
    newstring=Dd.sub("e",numberstring)
    return newstring
    # end convertd2e ==================================================


#======================================================================
def datafile2array (datafile=" ",sep=None, dtype="float",skiplines=0, \
    skipfirstcols=0, skiplastcols=0):
    """
        open and read data from a ascii text file into a
        numpy array. The number of rows and columns in the data file will match
        the size of the array.

        sep is the character seperation between fields in the data file
        dtype is the how the file data is to be interpretted.
        skiplines = n skips the first n line(s). 
        skipfirstcols = n skips the first n column(s)
        skiplastcols = n skips the last n column(s).
    """
    fid=open(datafile)
    data=fid.readlines()
    fid.close()

    dataarray=[]
    for row in range(skiplines,len(data)):
        data[row]=convertd2e(data[row])
        data[row]=string.split(data[row],sep)
        if data[row]!=[]:
            if dtype!=" ":
                for col in range(skipfirstcols,len(data[row])-skiplastcols) :
                    if dtype=="float":
                        data[row][col]=float(data[row][col])
                    elif dtype=="int":
                        data[row][col]=int(data[row][col])
            if dataarray!=[]: 
                if len(data[row])-skipfirstcols-skiplastcols==len(dataarray[0]):
                    dataarray.append(data[row][skipfirstcols:len(data[row])-skiplastcols])
            else:
                dataarray.append(data[row][skipfirstcols:len(data[row])-skiplastcols])
    
    dataarray=numpy.array(dataarray)
    return dataarray
    # end loaddatafile ==================================================

def array2datafile (dataarray,datafile=" ",sep=""):
    """
    def array2datafile (array,datafile=" "):

        output a tuple, list or numpy array into a ascii text file. 
        The number of rows and columns in the data file will match
        the size of the array.
   
    """
    fid=open(datafile,'w')
    shp=numpy.shape(dataarray)
    if len(shp)>1:
        for row in range(shp[0]):
            for col in range(shp[1]):
                fid.write("%s %s" % (dataarray[row][col],sep))

            fid.write("\n")
    else:
        for col in range(shp[0]) :
            fid.write("%s %s" % (dataarray[col],sep))

    fid.close()
    return
    # end array2datafile =====================================================
    




#!/usr/bin/env python
# encoding: utf-8
r"""
Data Module

Contains the general class definition and the subclasses of the Clawpack data
objects.

:Authors:
    Kyle T. Mandli and Randall J. LeVeque (2008-08-07) Initial version

    Randall J. LeVeque (2008-08-07) Plotting data objects

    Alan McIntyre (2009-01-01) Speed ups and rebuilding of Data

    Kyle T. Mandli (2009-04-01) Stripped down and improved version
"""
# ============================================================================
#      Copyright (C) 2008 Kyle T. Mandli <mandli@amath.washington.edu>
#      Copyright (C) 2008 Randall J. LeVeque <rjl@amath.washington.edu>
#      Copyright (C) 2009 Alan McIntyre <amcin001@amath.washington.edu>
#
#  Distributed under the terms of the Berkeley Software Distribution (BSD)
#  license
#                     http://www.opensource.org/licenses/
# ============================================================================

import shutil
import os
import copy
import re
import logging

# ========== Parse Value Utility Function ====================================
def _parse_value(value):
    r"""
    Attempt to make sense of a value string from a config file.  If the
    value is not obviously an integer, float, or boolean, it is returned as
    a string stripped of leading and trailing whitespace.

    :Input:
        - *value* - (string) Value string to be parsed

    :Output:
        - (id) - Appropriate object based on *value*
    """
    value = value.strip()
    if not value:
        return None

    # assume that values containing spaces are lists of values
    if len(value.split()) > 1:
        return [_parse_value(vv) for vv in value.split()]

    try:
        # see if it's an integer
        value = int(value)
    except ValueError:
        try:
            # see if it's a float
            value = float(value)
        except ValueError:
            # see if it's a bool
            if value[0] == 'T':
                value = True
            elif value[0] == 'F':
                value = False

    return value


# ============================================================================
#  General Data Class
# ============================================================================
class Data(object):
    r"""
    Generalized clawpack data object

    Generalized class for Clawpack data.  Contains generic methods for reading
    and writing data to and from a data file.

    :Initialization:
        Input:
         - *data_files* - (List of strings) Paths to data files to be read in,
           an empty data object can be created by providing no data files.
         - *attributes* - (List of strings) List of required attribute names
           which will be initialized to None.

    :Version: 1.2 (2009-04-01)
    """

    __name__ = 'Data'

    def __init__(self, data_files=[], attributes=None):
        """
        Initialize a Data object

        See :class:`Data` for more info.
        """

        # Internal bookkeeping variables
        self.__attributes = []
        self.__owners = {}

        # Setup data logger
        self.logger = logging.getLogger('data')

        # Initialize from attribute list provided
        if attributes:
            for attr in attributes:
                self.add_attribute(attr,None,None)

        # Read data files from data_files list
        if isinstance(data_files, str):
            data_files = [data_files]
        elif not isinstance(data_files, list):
            raise Exception("data_files must be a list of strings")

        if len(data_files) > 0:
            self.read(data_files)

    # ========== Return string representation of this Data Object ========
    def __str__(self):
        output = "%s%s%s\n" % ("Name".ljust(25),"Value".ljust(12),
                                    "Owner".ljust(12))
        for (k,v) in self.items():
            output += "%s%s%s\n" % (str(k).ljust(25),
                                    str(v).ljust(12),
                                    str(self.__owners[k]).ljust(12))
        return output

    # ========== Access Methods ==============================================
    def add_attribute(self, name, value=None, owner=None):
        r"""
        Adds an attribute called name to the data object

        If an attribute needs to be added to the object, this routine must be
        called or the attribute will not be written out.

        :Input:
         - *name* - (string) Name of the data attribute
         - *value* - (id) Value to set *name* to, defaults to None
         - *owner* - (id) Owner of this particular attribute
        """
        setattr(self,name,value)
        self.__owners[name] = owner
        if name not in self.__attributes:
            self.__attributes.append(name)

    def remove_attributes(self, arg_list):
        r"""
        Remove the listed attributes.
        """

        # Convert to list if args is not already a list
        if not isinstance(arg_list,list):
            arg_list = [arg_list]

        for arg in arg_list:
            self.__owners.pop(arg)
            self.__attributes.remove(arg)
            delattr(self,arg)

    def attributes():
        def fget(self): return self.__attributes
        return locals()
    attributes = property(**attributes())

    def has_attribute(self,name):
        r"""
        Check if this data object has the given attributes

        :Input:
         - *name* - (string) Name of attribute

        :Output:
         - (bool) - True if data object contains a data attribute name
        """
        return name in self.__attributes

    def set_owner(self,name,owner):
        r"""
        Sets the owner of the given data

        :Input:
         - *name* - (string) Name of attribute
         - *owner* - (id) Owner of the attribute
        """
        if name not in self.__attributes:
            raise KeyError("No attribute named %s" % name)
        self.__owners[name] = owner

    def get_owner(self,name):
        r"""
        Returns the owner of the data attribute name

        :Input:
         - *name* - (string) Name of attribute

        :Output:
         - (id) - Owner of attribute
        """
        return self.__owners[name]

    def get_owners(self,supplementary_file=None):
        r"""
        Returns a list of owners excluding the owner None

        If supplementary_file is provided, None is replace by that owner.

        :Input:
         - *supplementary_file* - (string) Supplementary file, defaults to
           None.

        :Output:
         - (list) - Returns a list of owners
        """
        owners = []
        for (key,owner) in self.__owners.items():
            if owner is None:
                self.__owners[key] = supplementary_file
            # This simultaneously finds one instance of an owner and tests
            # to see if the supplementary_file is not None
            owner = self.__owners[key]
            if owner not in owners and owner is not None:
                owners.append(owner)
        return owners

    def iteritems(self):
        r"""
        Returns an iterator of keys and values from this object

        :Output:
         - (Iterator) Iterator over keys and values
        """
        return [(k,getattr(self,k)) for k in self.__attributes]

    # ========== Read in a collection of data files ==========================
    def read(self,data_paths):
        r"""
        Read data in from a clawpack style data file.

        Any lines of the form::

           values  =:  name

        is used to set an attribute of self.

        INPUT

        data_paths : Path to a data file to be read in, can also be a list
                    of files to be read in.
        """
        if isinstance(data_paths, str):
            data_paths = [data_paths]

        for filename in data_paths:
            filename = os.path.abspath(filename)
            if not os.path.exists(filename):
                raise Exception("No such data file: %s" % filename)

            # self.logger.info("Reading from %s" % filename)

            for lineno, line in enumerate(open(filename)):
                if '=:' not in line:
                    continue

                value, tail = line.split('=:')
                varname = tail.split()[0]

                oldval = getattr(self, varname, None)
                newval = _parse_value(value)
                if oldval is not None:
                    vals = "(old=%r,new=%r)" % (oldval, newval)
                    # self.logger.debug("Overwriting %s %s" % (varname, vals))

                # if newval is None:
                    # self.logger.warning("Empty value for %s" % varname)

                self.add_attribute(varname, newval, filename)

    # ========== Write out the data from this object =========================
    def write(self,data_files=None,supplementary_file=None):
        r"""
        Write out the contents of the data object

        This method writes out the current required attributes of the data
        object to a file or list of files.  The format for the output will be
        in the form::

            values =: name

        The order is either retained from the files in the data list or
        written in the order they are in the list of attributes

        The behavior is determined by the arguments passed into the routine:

        No arguments:

            The contents of the object will be written out to the files listed
            in data_files only if each attribute is contained in the file.
            This implies that if an attribute is not located in any of the
            files in data_files then it will not be written out to file.

        If data_files is provided and data_files is a valid owner:

            Write out the attributes appropriate to that file

        If data_files is provided and not(data_files in owner):

            Write all attributes to the data_file given

        If supplementary_file is provided:

            Write out any attributes without an owner to this file, all owned
            attributes will be written out to the approriate files
        """

        # Expand supplementary_file if it is not None
        if supplementary_file is not None:
            supplementary_file = os.path.abspath(supplementary_file)

        # Create list of owners
        owners = self.get_owners(supplementary_file=supplementary_file)

        #print 'in write: data_files = ',data_files
        # Write to the entire owner list
        if data_files is None:
            file_list = owners
        elif isinstance(data_files,str):
            #print '<p>in write'; import sys; sys.exit(0)
            path = os.path.abspath(data_files)
            if path not in owners:
                # Create temporary data file to store all data in the object
                try:
                    temp_file = open(path,'w')
                    for key in self.__attributes:
                        temp_file.write("1 =: %s\n" % key)
                    temp_file.close()
                except IOError as xxx_todo_changeme:
                    (errno, strerror) = xxx_todo_changeme.args
                    print("I/O error(%s): %s" % (errno, strerror))
                    raise
                except:
                    raise
            # Add the path to the file_list
            file_list = [path]
        # Check to make sure all the paths are in the owner list
        elif isinstance(data_files,list):
            for path in data_files:
                path = os.path.abspath(path)
                if not(path in owners):
                    print("%s is not a registered owner!")
                    return
            file_list = data_files
        else:
            raise Exception("Invalid argument list given to write().")

        # Create temporary supplementary file if requested
        if supplementary_file is not None:
            try:
                sup_file = open(supplementary_file,'w')
                for attr in self.__attributes:
                    if self.__owners[attr] is supplementary_file:
                        sup_file.write("-1 =: %s\n" % attr)
                        self.__owners[attr] = supplementary_file
                sup_file.close()
            except:
                raise


        # Regular expression for searching each file
        regexp = re.compile(r"(?P<values>.*)=:(?P<name>.*)")
        # Loop over each file
        for data_path in file_list:
            # Open the data file and temporary file
            try:
                data_file = open(data_path,'r')
            except(IOError):
                raise
            try:
                temp_path = os.path.join(os.path.dirname(data_path), \
                    'temp.' + os.path.basename(data_path))
                temp_file = open(temp_path,'w')
            except(IOError):
                print("IOERROR")
                raise

            try:
                for line in data_file:
                    result = regexp.search(line)
                    if result:
                        name = re.split(r'\s+', result.group('name').strip())[0]
                        values = re.split(r'\s+', result.group('values').strip())

                        if len(values) == 0:
                            line = ''
                        elif self.__owners[name] == data_path or \
                                data_path not in self.__owners:
                            newvalues = getattr(self,name)

                            # Convert newvalues to an appropriate string repr
                            if isinstance(newvalues,tuple) \
                                | isinstance(newvalues,list):
                                # Remove [], (), and ','
                                newstring = repr(newvalues)[1:-1]
                                newstring = newstring.replace(',','')
                            elif isinstance(newvalues,bool):
                                if newvalues:
                                    newstring = 'T'
                                else:
                                    newstring = 'F'
                            else:
                                newstring = repr(newvalues)

                            newstart = newstring.ljust(25)
                            line = line.replace(result.group('values') + "=:", \
                                newstart + " =:")
                        else:
                            print("Error writing out %s" % name)
                            raise AttributeError(name)

                    # Write the new line
                    temp_file.write(line)
            except:
                raise

            # Close files
            data_file.close()
            temp_file.close()

            # Rename the temporary file to the data_path name
            try:
                shutil.move(temp_path,data_path)
            except:
                raise


#-----------------------------------------------------------------------

# New classes and functions for dealing with data in setrun function.

class ClawInputData(Data):
    r"""
    Object that will be written out to claw.data.
    """
    def __init__(self, ndim):
        super(ClawInputData,self).__init__()
        self.add_attribute('ndim',ndim)

        # Set default values:
        if ndim == 1:
            self.add_attribute('mx',100)
            self.add_attribute('nout',5)
            self.add_attribute('outstyle',1)
            self.add_attribute('tfinal',1.0)
            self.add_attribute('dt_initial',1.e-5)
            self.add_attribute('dt_max',1.e99)
            self.add_attribute('cfl_desired',0.9)
            self.add_attribute('cfl_max',1.0)
            self.add_attribute('max_steps',5000)
            self.add_attribute('dt_variable',1)
            self.add_attribute('order',2)
            self.add_attribute('order_trans',0)
            self.add_attribute('verbosity',0)
            self.add_attribute('src_split',0)
            self.add_attribute('mcapa',0)
            self.add_attribute('maux',0)
            self.add_attribute('meqn',1)
            self.add_attribute('mwaves',1)
            self.add_attribute('mthlim',[4])
            self.add_attribute('t0',0.)
            self.add_attribute('xlower',0.)
            self.add_attribute('xupper',1.)
            self.add_attribute('mbc',2)
            self.add_attribute('mthbc_xlower',1)
            self.add_attribute('mthbc_xupper',1)
            self.add_attribute('restart',0)
            self.add_attribute('N_restart',0)


        elif ndim == 2:
            self.add_attribute('mx',100)
            self.add_attribute('my',100)
            self.add_attribute('nout',5)
            self.add_attribute('outstyle',1)
            self.add_attribute('tfinal',1.0)
            self.add_attribute('dt_initial',1.e-5)
            self.add_attribute('dt_max',1.e99)
            self.add_attribute('cfl_desired',0.9)
            self.add_attribute('cfl_max',1.0)
            self.add_attribute('max_steps',5000)
            self.add_attribute('dt_variable',1)
            self.add_attribute('order',2)
            self.add_attribute('order_trans',2)
            self.add_attribute('verbosity',0)
            self.add_attribute('src_split',0)
            self.add_attribute('mcapa',0)
            self.add_attribute('maux',0)
            self.add_attribute('meqn',1)
            self.add_attribute('mwaves',1)
            self.add_attribute('mthlim',[4])
            self.add_attribute('t0',0.)
            self.add_attribute('xlower',0.)
            self.add_attribute('xupper',1.)
            self.add_attribute('ylower',0.)
            self.add_attribute('yupper',1.)
            self.add_attribute('mbc',2)
            self.add_attribute('mthbc_xlower',1)
            self.add_attribute('mthbc_xupper',1)
            self.add_attribute('mthbc_ylower',1)
            self.add_attribute('mthbc_yupper',1)
            self.add_attribute('restart',0)
            self.add_attribute('N_restart',0)

        else:
            raise AttributeError("Only ndim=1 or 2 supported so far")

    def write(self):
        print('Creating data file claw.data for use with xclaw')
        make_clawdatafile(self)



class AmrclawInputData(Data):
    r"""
    Object that will be written out to amr2ez.data.
    """
    def __init__(self, ndim):
        super(AmrclawInputData,self).__init__()
        self.add_attribute('ndim',ndim)

        # Set default values:
        if ndim == 1:
            self.add_attribute('mx',100)
            self.add_attribute('nout',5)
            self.add_attribute('outstyle',1)
            self.add_attribute('tfinal',1.0)
            self.add_attribute('dt_initial',1.e-5)
            self.add_attribute('dt_max',1.e99)
            self.add_attribute('cfl_desired',0.9)
            self.add_attribute('cfl_max',1.0)
            self.add_attribute('max_steps',5000)
            self.add_attribute('dt_variable',1)
            self.add_attribute('order',2)
            self.add_attribute('order_trans',0)
            self.add_attribute('verbosity',0)
            self.add_attribute('src_split',0)
            self.add_attribute('mcapa',0)
            self.add_attribute('maux',0)
            self.add_attribute('meqn',1)
            self.add_attribute('mwaves',1)
            self.add_attribute('mthlim',[4])
            self.add_attribute('t0',0.)
            self.add_attribute('xlower',0.)
            self.add_attribute('xupper',1.)
            self.add_attribute('mbc',2)
            self.add_attribute('mthbc_xlower',1)
            self.add_attribute('mthbc_xupper',1)
            self.add_attribute('restart',0)
            self.add_attribute('N_restart',0)

            # attributes need only since AMR is done using 2d amrclaw:
            self.add_attribute('my',1)
            self.add_attribute('ylower',0.)
            self.add_attribute('yupper',1.)
            self.add_attribute('mthbc_ylower',1)
            self.add_attribute('mthbc_yupper',1)
            self.add_attribute('inraty',[1,1,1,1,1,1])

        elif ndim == 2:
            self.add_attribute('mx',100)
            self.add_attribute('my',100)
            self.add_attribute('nout',5)
            self.add_attribute('outstyle',1)
            self.add_attribute('tfinal',1.0)
            self.add_attribute('dt_initial',1.e-5)
            self.add_attribute('dt_max',1.e99)
            self.add_attribute('cfl_desired',0.9)
            self.add_attribute('cfl_max',1.0)
            self.add_attribute('max_steps',5000)
            self.add_attribute('dt_variable',1)
            self.add_attribute('order',2)
            self.add_attribute('order_trans',2)
            self.add_attribute('verbosity',0)
            self.add_attribute('src_split',0)
            self.add_attribute('mcapa',0)
            self.add_attribute('maux',0)
            self.add_attribute('meqn',1)
            self.add_attribute('mwaves',1)
            self.add_attribute('mthlim',[4])
            self.add_attribute('t0',0.)
            self.add_attribute('xlower',0.)
            self.add_attribute('xupper',1.)
            self.add_attribute('ylower',0.)
            self.add_attribute('yupper',1.)
            self.add_attribute('mbc',2)
            self.add_attribute('mthbc_xlower',1)
            self.add_attribute('mthbc_xupper',1)
            self.add_attribute('mthbc_ylower',1)
            self.add_attribute('mthbc_yupper',1)
            self.add_attribute('restart',0)
            self.add_attribute('N_restart',0)
            self.add_attribute('inraty',[1])

        if ndim <= 2:
            # AMR parameters:
            self.add_attribute('mxnest',-1)
            self.add_attribute('inratx',[1])
            self.add_attribute('inratt',[1])
            self.add_attribute('auxtype',[])
            self.add_attribute('restart',False)
            self.add_attribute('checkpt_iousr',1000)
            self.add_attribute('tchk',[])
            self.add_attribute('tol',-1.0)
            self.add_attribute('tolsp',0.05)
            self.add_attribute('kcheck',2)
            self.add_attribute('ibuff',3)
            self.add_attribute('cutoff',0.7)
            self.add_attribute('PRINT',False)
            self.add_attribute('NCAR',False)
            self.add_attribute('fortq',True)
            self.add_attribute('dprint',False)
            self.add_attribute('eprint',False)
            self.add_attribute('edebug',False)
            self.add_attribute('gprint',False)
            self.add_attribute('nprint',False)
            self.add_attribute('pprint',False)
            self.add_attribute('rprint',False)
            self.add_attribute('sprint',False)
            self.add_attribute('tprint',False)
            self.add_attribute('uprint',False)
        else:
            print('*** Error: only ndim=1 or 2 supported so far ***')
            raise AttributeError("Only ndim=1 or 2 supported so far")

    def write(self):
        print('Creating data file amr2ez.data for use with xamr')
        make_amrclawdatafile(self)
        make_setgauges_datafile(self)


class SharpclawInputData(Data):
    r"""
    Object that will be written out to claw.data.
    """
    def __init__(self, ndim):
        super(SharpclawInputData,self).__init__()
        self.add_attribute('ndim',ndim)

        # Set default values:
        if ndim == 1:
            self.add_attribute('mx',100)
            self.add_attribute('nout',5)
            self.add_attribute('outstyle',1)
            self.add_attribute('tfinal',1.0)
            self.add_attribute('dt_initial',1.e-5)
            self.add_attribute('dt_max',1.e99)
            self.add_attribute('cfl_desired',0.9)
            self.add_attribute('cfl_max',1.0)
            self.add_attribute('max_steps',5000)
            self.add_attribute('dt_variable',1)
            self.add_attribute('verbosity',0)
            self.add_attribute('mcapa',0)
            self.add_attribute('maux',0)
            self.add_attribute('meqn',1)
            self.add_attribute('mwaves',1)
            self.add_attribute('mthlim',[5])
            self.add_attribute('t0',0.)
            self.add_attribute('xlower',0.)
            self.add_attribute('xupper',1.)
            self.add_attribute('mbc',3)
            self.add_attribute('mthbc_xlower',1)
            self.add_attribute('mthbc_xupper',1)
            self.add_attribute('restart',0)
            self.add_attribute('N_restart',0)
            self.add_attribute('time_integrator',2)
            self.add_attribute('tfluct_solver',0)
            self.add_attribute('char_decomp',0)
            self.add_attribute('lim_type',2)
            self.add_attribute('src_term',0)


        elif ndim == 2:
            self.add_attribute('mx',100)
            self.add_attribute('my',100)
            self.add_attribute('nout',5)
            self.add_attribute('outstyle',1)
            self.add_attribute('tfinal',1.0)
            self.add_attribute('dt_initial',1.e-5)
            self.add_attribute('dt_max',1.e99)
            self.add_attribute('cfl_desired',0.9)
            self.add_attribute('cfl_max',1.0)
            self.add_attribute('max_steps',5000)
            self.add_attribute('dt_variable',1)
            self.add_attribute('verbosity',0)
            self.add_attribute('mcapa',0)
            self.add_attribute('maux',0)
            self.add_attribute('meqn',1)
            self.add_attribute('mwaves',1)
            self.add_attribute('mthlim',[5])
            self.add_attribute('t0',0.)
            self.add_attribute('xlower',0.)
            self.add_attribute('xupper',1.)
            self.add_attribute('ylower',0.)
            self.add_attribute('yupper',1.)
            self.add_attribute('mbc',3)
            self.add_attribute('mthbc_xlower',1)
            self.add_attribute('mthbc_xupper',1)
            self.add_attribute('mthbc_ylower',1)
            self.add_attribute('mthbc_yupper',1)
            self.add_attribute('restart',0)
            self.add_attribute('N_restart',0)
            self.add_attribute('time_integrator',2)
            self.add_attribute('tfluct_solver',0)
            self.add_attribute('char_decomp',0)
            self.add_attribute('lim_type',2)
            self.add_attribute('src_term',0)

        else:
            raise AttributeError("Only ndim=1 or 2 supported so far")

    def write(self):
        print('Creating data file sharpclaw.data for use with xsclaw')
        make_sharpclawdatafile(self)




def open_datafile(name, datasource='setrun.py'):
    """
    Open a data file and write a warning header.
    Warning header starts with '#' character.  These lines are skipped if
    data file is opened using the library routine opendatafile.

    :Input:
     - *name* - (string) Name of data file
     - *datasource* - (string) Source for the data

    :Output:
     - (file) - file object
    """

    import string

    source = datasource.ljust(25)
    file = open(name, 'w')
    file.write('########################################################\n')
    file.write('### DO NOT EDIT THIS FILE:  GENERATED AUTOMATICALLY ####\n')
    file.write('### To modify data, edit  %s ####\n' % source)
    file.write('###    and then "make .data"                        ####\n')
    file.write('########################################################\n\n')

    return file


def data_write(file, dataobj, name=None, descr=''):
    r"""
    Write out value to data file, in the form ::

       value =: name  descr

    Remove brackets and commas from lists, and replace booleans by T/F.
    Also convert numpy array to a list first.

    :Input:
     - *name* - (string) normally a string defining the variable,
       ``if name==None``, write a blank line.
     - *descr* - (string) A short description to appear on the line
    """

    import string
    if name is None:
        file.write('\n')
    else:
        try:
            value = getattr(dataobj, name)
        except:
            print("Variable missing: ",name)
            print("  from dataobj = ", dataobj)
            raise
        # Convert value to an appropriate string repr
        import numpy 
        if isinstance(value,numpy.ndarray):
            value = list(value)
        if isinstance(value,tuple) | isinstance(value,list):
            # Remove [], (), and ','
            string_value = repr(value)[1:-1]
            string_value = string_value.replace(',','')
        elif isinstance(value,bool):
            if value:
                string_value = 'T'
            else:
                string_value = 'F'
        else:
            string_value = repr(value)
        padded_value = string_value.ljust(25)
        padded_name = name.ljust(12)
        file.write('%s =: %s %s\n' % (padded_value, padded_name, descr))


def make_clawdatafile(clawdata):
    r"""
    Take the data specified in clawdata and write it to claw.data in the
    form required by the Fortran code lib/main.f95.
    """


    # open file and write a warning header:
    file = open_datafile('claw.data')

    ndim = clawdata.ndim
    data_write(file, clawdata, 'ndim', '(number of dimensions)')
    data_write(file, clawdata, 'mx', '(cells in x direction)')
    if ndim > 1:
        data_write(file, clawdata, 'my', '(cells in y direction)')
    if ndim == 3:
        data_write(file, clawdata, 'mz', '(cells in z direction)')
    data_write(file, clawdata, None)  # writes blank line

    data_write(file, clawdata, 'nout', '(number of output times)')
    data_write(file, clawdata, 'outstyle', '(style of specifying output times)')
    if clawdata.outstyle == 1:
        data_write(file, clawdata, 'tfinal', '(final time)')
    elif clawdata.outstyle == 2:
        data_write(file, clawdata, 'tout', '(output times)')
    elif clawdata.outstyle == 3:
        data_write(file, clawdata, 'iout', '(output every iout steps)')
    elif clawdata.outstyle == 4:
        data_write(file, clawdata, 'output_time_interval', '(between outputs)')
        data_write(file, clawdata, 'tfinal', '(final time)')

    else:
        print('*** Error: unrecognized outstyle')
        raise
        return

    data_write(file, clawdata, None)
    data_write(file, clawdata, 'dt_initial', '(initial time step dt)')
    data_write(file, clawdata, 'dt_max', '(max allowable dt)')
    data_write(file, clawdata, 'cfl_max', '(max allowable Courant number)')
    data_write(file, clawdata, 'cfl_desired', '(desired Courant number)')
    data_write(file, clawdata, 'max_steps', '(max time steps per call to claw)')
    data_write(file, clawdata, None)
    data_write(file, clawdata, 'dt_variable', '(1 for variable dt, 0 for fixed)')
    data_write(file, clawdata, 'order', '(1 or 2)')
    if ndim == 1:
        data_write(file, clawdata, 'order_trans', '(not used in 1d)')
    else:
        data_write(file, clawdata, 'order_trans', '(transverse order)')
    data_write(file, clawdata, 'verbosity', '(verbosity of output)')
    data_write(file, clawdata, 'src_split', '(source term splitting)')
    data_write(file, clawdata, 'mcapa', '(aux index for capacity fcn)')
    data_write(file, clawdata, 'maux', '(number of aux variables)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 'meqn', '(number of equations)')
    data_write(file, clawdata, 'mwaves', '(number of waves)')
    data_write(file, clawdata, 'mthlim', '(limiter choice for each wave)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 't0', '(initial time)')
    data_write(file, clawdata, 'xlower', '(xlower)')
    data_write(file, clawdata, 'xupper', '(xupper)')
    if ndim > 1:
        data_write(file, clawdata, 'ylower', '(ylower)')
        data_write(file, clawdata, 'yupper', '(yupper)')
    if ndim == 3:
        data_write(file, clawdata, 'zlower', '(zlower)')
        data_write(file, clawdata, 'zupper', '(zupper)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 'mbc', '(number of ghost cells)')
    data_write(file, clawdata, 'mthbc_xlower', '(type of BC at xlower)')
    data_write(file, clawdata, 'mthbc_xupper', '(type of BC at xupper)')
    if ndim > 1:
        data_write(file, clawdata, 'mthbc_ylower', '(type of BC at ylower)')
        data_write(file, clawdata, 'mthbc_yupper', '(type of BC at yupper)')
    if ndim == 3:
        data_write(file, clawdata, 'mthbc_zlower', '(type of BC at zlower)')
        data_write(file, clawdata, 'mthbc_zupper', '(type of BC at zupper)')

    data_write(file, clawdata, 'restart', '(1 to restart from a past run)')
    data_write(file, clawdata, 'N_restart', '(which frame to restart from)')
    data_write(file, clawdata, None)

    file.close()

def make_amrclawdatafile(clawdata):
    r"""
    Take the data specified in clawdata and write it to claw.data in the
    form required by the Fortran code lib/main.f95.
    """


    # open file and write a warning header:
    file = open_datafile('amr2ez.data')

    ndim = clawdata.ndim
    #data_write(file, clawdata, 'ndim', '(number of dimensions)')
    data_write(file, clawdata, 'mx', '(cells in x direction)')
    data_write(file, clawdata, 'my', '(cells in y direction)')
    if ndim == 3:
        data_write(file, clawdata, 'mz', '(cells in z direction)')

    data_write(file, clawdata, 'mxnest', '(max number of grid levels)')
    if len(clawdata.inratx) < max(abs(clawdata.mxnest)-1, 1):
        raise ValueError("*** Error in data parameter: " + \
              "require len(inratx) >= %s " % max(abs(clawdata.mxnest) - 1, 1))
    data_write(file, clawdata, 'inratx', '(refinement ratios)')
    if clawdata.mxnest < 0:
        # negative mxnest indicates anisotropic refinement
        if len(clawdata.inraty) < max(abs(clawdata.mxnest)-1, 1):
            raise ValueError("*** Error in data parameter: " + \
              "require len(inraty) >= %s " % max(abs(clawdata.mxnest) - 1, 1))
        data_write(file, clawdata, 'inraty', '(refinement ratios)')
        if ndim == 3:
            if len(clawdata.inratz) < max(abs(clawdata.mxnest)-1, 1):
                raise ValueError("*** Error in data parameter: " + \
                  "require len(inratz) >= %s " % max(abs(clawdata.mxnest) - 1, 1))
            data_write(file, clawdata, 'inratz', '(refinement ratios)')
        if len(clawdata.inratt) < max(abs(clawdata.mxnest)-1, 1):
            raise ValueError("*** Error in data parameter: " + \
                  "require len(inratt) >= %s " % max(abs(clawdata.mxnest) - 1, 1))
        data_write(file, clawdata, 'inratt', '(refinement ratios)')

    data_write(file, clawdata, None)  # writes blank line

    data_write(file, clawdata, 'nout', '(number of output times)')
    data_write(file, clawdata, 'outstyle', '(style of specifying output times)')
    if clawdata.outstyle == 1:
        data_write(file, clawdata, 'tfinal', '(final time)')
    elif clawdata.outstyle == 2:
        data_write(file, clawdata, 'tout', '(output times)')
    elif clawdata.outstyle == 3:
        data_write(file, clawdata, 'iout', '(output every iout steps)')
    elif clawdata.outstyle == 4:
        data_write(file, clawdata, 'output_time_interval', '(between outputs)')
        data_write(file, clawdata, 'tfinal', '(final time)')
    else:
        print('*** Error: unrecognized outstyle')
        raise
        return

    data_write(file, clawdata, None)
    data_write(file, clawdata, 'dt_initial', '(initial time step dt)')
    data_write(file, clawdata, 'dt_max', '(max allowable dt)')
    data_write(file, clawdata, 'cfl_max', '(max allowable Courant number)')
    data_write(file, clawdata, 'cfl_desired', '(desired Courant number)')
    data_write(file, clawdata, 'max_steps', '(max time steps per call to claw)')
    data_write(file, clawdata, None)
    data_write(file, clawdata, 'dt_variable', '(1 for variable dt, 0 for fixed)')
    data_write(file, clawdata, 'order', '(1 or 2)')
    if ndim == 1:
        data_write(file, clawdata, 'order_trans', '(not used in 1d)')
    else:
        data_write(file, clawdata, 'order_trans', '(transverse order)')
    data_write(file, clawdata, 'verbosity', '(verbosity of output)')
    data_write(file, clawdata, 'src_split', '(source term splitting)')
    data_write(file, clawdata, 'mcapa', '(aux index for capacity fcn)')
    data_write(file, clawdata, 'maux', '(number of aux variables)')
    if len(clawdata.auxtype) != clawdata.maux:
        file.close()
        print("*** Error: An auxtype array must be specified of length maux")
        raise AttributeError("require len(clawdata.auxtype) == clawdata.maux")
    for i in range(clawdata.maux):
        file.write("'%s'\n" % clawdata.auxtype[i])
    data_write(file, clawdata, None)

    data_write(file, clawdata, 'meqn', '(number of equations)')
    data_write(file, clawdata, 'mwaves', '(number of waves)')
    data_write(file, clawdata, 'mthlim', '(limiter choice for each wave)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 't0', '(initial time)')
    data_write(file, clawdata, 'xlower', '(xlower)')
    data_write(file, clawdata, 'xupper', '(xupper)')
    data_write(file, clawdata, 'ylower', '(ylower)')
    data_write(file, clawdata, 'yupper', '(yupper)')
    if ndim == 3:
        data_write(file, clawdata, 'zlower', '(zlower)')
        data_write(file, clawdata, 'zupper', '(zupper)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 'mbc', '(number of ghost cells)')
    data_write(file, clawdata, 'mthbc_xlower', '(type of BC at xlower)')
    data_write(file, clawdata, 'mthbc_xupper', '(type of BC at xupper)')
    data_write(file, clawdata, 'mthbc_ylower', '(type of BC at ylower)')
    data_write(file, clawdata, 'mthbc_yupper', '(type of BC at yupper)')
    if ndim == 3:
        data_write(file, clawdata, 'mthbc_zlower', '(type of BC at zlower)')
        data_write(file, clawdata, 'mthbc_zupper', '(type of BC at zupper)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 'restart', '(1 to restart from a past run)')
    data_write(file, clawdata, 'checkpt_iousr', '(how often to checkpoint)')
    if clawdata.checkpt_iousr < 0:
        data_write(file, clawdata, 'tchk', '(checkpoint times)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 'tol', '(tolerance for Richardson extrap)')
    data_write(file, clawdata, 'tolsp', '(tolerance used in flag2refine)')
    data_write(file, clawdata, 'kcheck', '(how often to regrid)')
    data_write(file, clawdata, 'ibuff', '(buffer zone around flagged pts)')
    data_write(file, clawdata, 'cutoff', '(efficiency cutoff for grid gen.)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 'PRINT', '(print to fort.amr)')
    data_write(file, clawdata, 'NCAR', '(obsolete!)')
    data_write(file, clawdata, 'fortq', '(Output to fort.q* files)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 'dprint', '(print domain flags)')
    data_write(file, clawdata, 'eprint', '(print err est flags)')
    data_write(file, clawdata, 'edebug', '(even more err est flags)')
    data_write(file, clawdata, 'gprint', '(grid bisection/clustering)')
    data_write(file, clawdata, 'nprint', '(proper nesting output)')
    data_write(file, clawdata, 'pprint', '(proj. of tagged points)')
    data_write(file, clawdata, 'rprint', '(print regridding summary)')
    data_write(file, clawdata, 'sprint', '(space/memory output)')
    data_write(file, clawdata, 'tprint', '(time step reporting each level)')
    data_write(file, clawdata, 'uprint', '(update/upbnd reporting)')

    file.close()


def make_sharpclawdatafile(clawdata):
    r"""
    Take the data specified in clawdata and write it to sharpclaw.data in the
    form required by the Fortran code lib/main.f95.
    """


    # open file and write a warning header:
    file = open_datafile('sharpclaw.data')

    ndim = clawdata.ndim
    data_write(file, clawdata, 'ndim', '(number of dimensions)')
    data_write(file, clawdata, 'mx', '(cells in x direction)')
    if ndim > 1:
        data_write(file, clawdata, 'my', '(cells in y direction)')
    if ndim == 3:
        data_write(file, clawdata, 'mz', '(cells in z direction)')
    data_write(file, clawdata, None)  # writes blank line

    data_write(file, clawdata, 'nout', '(number of output times)')
    data_write(file, clawdata, 'outstyle', '(style of specifying output times)')
    if clawdata.outstyle == 1:
        data_write(file, clawdata, 'tfinal', '(final time)')
    elif clawdata.outstyle == 2:
        data_write(file, clawdata, 'tout', '(output times)')
    elif clawdata.outstyle == 3:
        data_write(file, clawdata, 'iout', '(output every iout steps)')
    else:
        print('*** Error: unrecognized outstyle')
        raise
        return

    data_write(file, clawdata, None)
    data_write(file, clawdata, 'dt_initial', '(initial time step dt)')
    data_write(file, clawdata, 'dt_max', '(max allowable dt)')
    data_write(file, clawdata, 'cfl_max', '(max allowable Courant number)')
    data_write(file, clawdata, 'cfl_desired', '(desired Courant number)')
    data_write(file, clawdata, 'max_steps', '(max time steps per call to claw)')
    data_write(file, clawdata, None)
    data_write(file, clawdata, 'dt_variable', '(1 for variable dt, 0 for fixed)')
    data_write(file, clawdata, 'time_integrator', '(time stepping scheme)')
    data_write(file, clawdata, 'verbosity', '(verbosity of output)')
    data_write(file, clawdata, 'src_term', '(source term present)')
    data_write(file, clawdata, 'mcapa', '(aux index for capacity fcn)')
    data_write(file, clawdata, 'maux', '(number of aux variables)')
    data_write(file, clawdata, 'tfluct_solver', '(total fluctuation solver)')
    data_write(file, clawdata, 'char_decomp', '(characteristic decomposition)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 'meqn', '(number of equations)')
    data_write(file, clawdata, 'mwaves', '(number of waves)')
    data_write(file, clawdata, 'lim_type', '(0=None, 1=TVD, 2=WENO)')
    data_write(file, clawdata, 'mthlim', '(limiter choice for each wave)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 't0', '(initial time)')
    data_write(file, clawdata, 'xlower', '(xlower)')
    data_write(file, clawdata, 'xupper', '(xupper)')
    if ndim > 1:
        data_write(file, clawdata, 'ylower', '(ylower)')
        data_write(file, clawdata, 'yupper', '(yupper)')
    if ndim == 3:
        data_write(file, clawdata, 'zlower', '(zlower)')
        data_write(file, clawdata, 'zupper', '(zupper)')
    data_write(file, clawdata, None)

    data_write(file, clawdata, 'mbc', '(number of ghost cells)')
    data_write(file, clawdata, 'mthbc_xlower', '(type of BC at xlower)')
    data_write(file, clawdata, 'mthbc_xupper', '(type of BC at xupper)')
    if ndim > 1:
        data_write(file, clawdata, 'mthbc_ylower', '(type of BC at ylower)')
        data_write(file, clawdata, 'mthbc_yupper', '(type of BC at yupper)')
    if ndim == 3:
        data_write(file, clawdata, 'mthbc_zlower', '(type of BC at zlower)')
        data_write(file, clawdata, 'mthbc_zupper', '(type of BC at zupper)')

    data_write(file, clawdata, 'restart', '(1 to restart from a past run)')
    data_write(file, clawdata, 'N_restart', '(which frame to restart from)')
    data_write(file, clawdata, None)

    file.close()


def make_userdatafile(userdata):
    r"""
    Create the data file using the parameters in userdata.
    The parameters will be written to this file in the same order they were
    specified using userdata.add_attribute.
    Presumably the user will read these in using a Fortran routine, such as
    setprob.f95, and the order is important.
    """

    # open file and write a warning header:
    file = open_datafile(userdata._UserData__fname)

    # write all the parameters:
    for param in userdata.attributes:
        data_write(file, userdata, param, \
                   userdata._UserData__descr[param])

    file.close()

def make_setgauges_datafile(clawdata):
    """
    Create setgauges.data using gauges attribute of clawdata.
    """
    gauges = getattr(clawdata,'gauges',[])
    ngauges = len(gauges)

    print('Creating data file setgauges.data')
    # open file and write a warning header:
    file = open_datafile('setgauges.data')
    file.write("%4i   =: ngauges\n" % ngauges)
    gaugeno_used = []
    for gauge in gauges:
        gaugeno = gauge[0]
        if gaugeno in gaugeno_used:
            print("*** Gauge number %s used more than once! " % gaugeno)
            raise Exception("Repeated gauge number")
        else:
            gaugeno_used.append(gauge[0])
        file.write("%4i %19.10e  %17.10e  %13.6e  %13.6e\n" % tuple(gauge))
        # or use this variant with =:
        #gauge.append(gaugeno)
        #file.write("%4i %19.10e  %17.10e  %13.6e  %13.6e  =: gauge%s\n" % tuple(gauge))
    file.close()


#-----------------------------------------------------
# New version 6/30/09

class ClawRunData(Data):
    r"""
    Object that will be written out to claw.data.
    """
    def __init__(self, pkg, ndim):
        super(ClawRunData,self).__init__()
        self.add_attribute('pkg',pkg)
        self.add_attribute('ndim',ndim)
        self.add_attribute('datalist',[])


        if pkg.lower() in ['classic', 'classicclaw']:
            self.add_attribute('xclawcmd', 'xclaw')

            # Required data set for basic run parameters:
            clawdata = ClawInputData(ndim)
            self.add_attribute('clawdata', clawdata)
            self.datalist.append(clawdata)

        elif pkg.lower() in ['amrclaw', 'amr']:
            self.add_attribute('xclawcmd', 'xamr')

            # Required data set for basic run parameters:
            clawdata = AmrclawInputData(ndim)
            self.add_attribute('clawdata', clawdata)
            self.datalist.append(clawdata)

        elif pkg.lower() in ['geoclaw']:
            self.add_attribute('xclawcmd', 'xgeoclaw')

            # Required data set for basic run parameters:
            clawdata = AmrclawInputData(ndim)
            self.add_attribute('clawdata', clawdata)
            self.datalist.append(clawdata)
            geodata = GeoclawInputData(ndim)
            self.add_attribute('geodata', geodata)
            self.datalist.append(geodata)

        elif pkg.lower() in ['sharpclaw']:
            self.add_attribute('xclawcmd', 'xsclaw')

            # Required data set for basic run parameters:
            clawdata = SharpclawInputData(ndim)
            self.add_attribute('clawdata', clawdata)
            self.datalist.append(clawdata)

        else:
            raise AttributeError("Unrecognized Clawpack pkg = %s" % pkg)

    def new_UserData(self,name,fname):
        r"""
        Create a new attribute called name
        for application specific data to be written
        to the data file fname.
        """
        userdata = UserData(fname)
        self.datalist.append(userdata)
        exec('self.%s = userdata' % name)
        return userdata

    def add_GaugeData(self):
        r"""
        Create a gaugedata attribute for writing to gauges.data.
        """
        gaugedata = GaugeData(self.ndim)
        self.datalist.append(gaugedata)
        self.gaugedata = gaugedata
        return gaugedata

    def write(self):
        for d in self.datalist:
            d.write()

class UserData(Data):
    r"""
    Object that will be written out to user file such as setprob.data, as
    determined by the fname attribute.
    """
    def __init__(self, fname):
        super(UserData,self).__init__()
        self.__fname = fname  # file to be read by Fortran for this data
        self.__descr = {}     # dictionary to hold descriptions

    def add_param(self,name,value,descr=''):
         self.add_attribute(name,value)
         self.__descr[name] = descr

    def write(self):
         print('Creating data file %s' % self.__fname)
         make_userdatafile(self)

class GaugeData(Data):
    r"""
    Data to be written out to gauge.data specifying gauges.
    DEPRECATED:  Use GeoclawInputData  instead.
    """
    def __init__(self, ndim):
        super(GaugeData,self).__init__()
        self.add_attribute('ndim',ndim)
        self.add_attribute('ngauges',0)
        self.__gauge_dict = {}

    def add_gauge(self,gaugeno,location,time_interval):
        self.__gauge_dict[gaugeno] = (gaugeno, location, time_interval)
        self.ngauges = len(self.__gauge_dict)

    def write(self):
        print('Creating data file gauges.data')

        # open file and write a warning header:
        file = open_datafile('gauges.data')

        data_write(file, self, 'ngauges', 'Number of gauges')
        data_write(file, self, None)

        ndim = self.ndim

        # write a line for each gauge:
        for (gaugeno, gdata) in self.__gauge_dict.items():
            tmin = gdata[2][0]
            tmax = gdata[2][1]
            if isinstance(gdata[1],(list,tuple)):
                xyz = gdata[1]
                x = xyz[0]
                if ndim>1:
                    y = xyz[1]
                if ndim>2:
                    z = xyz[2]
            else:
                x = gdata[1]

            if ndim==1:
                file.write('%i   %e   %e   %e' % (gdata[0],x,tmin,tmax))
            elif ndim==2:
                file.write('%i   %e   %e   %e   %e' % (gdata[0],x,y,tmin,tmax))
            elif ndim==3:
                file.write('%i   %e   %e   %e   %e   %e' % (gdata[0],x,y,z,tmin,tmax))

        printxyz = {1: 'x  ', 2: 'x  y  ', 3: 'x  y  z'}
        file.write('\n\n# Format of each line: \n#   gaugeno  %s tmin tmax'\
                     % printxyz[ndim])
        file.close()

class GeoclawInputData(Data):
    r"""
    Object that will be written out to the various GeoClaw data files.
    """
    def __init__(self, ndim):
        super(GeoclawInputData,self).__init__()

        # Set default values:
        self.add_attribute('igravity',1)
        self.add_attribute('iqinit',0)
        self.add_attribute('icoriolis',1)
        self.add_attribute('Rearth',6367500.0)
        self.add_attribute('variable_dt_refinement_ratios',False)
        # NEED TO CONTINUE!

    def write(self):

        print('Creating data file setgeo.data')
        # open file and write a warning header:
        file = open_datafile('setgeo.data')
        data_write(file, self, 'igravity')
        data_write(file, self, 'gravity')
        data_write(file, self, 'icoordsys')
        data_write(file, self, 'icoriolis')
        data_write(file, self, 'Rearth')
        data_write(file, self, 'variable_dt_refinement_ratios')
        file.close()

        print('Creating data file settsunami.data')
        # open file and write a warning header:
        file = open_datafile('settsunami.data')
        data_write(file, self, 'sealevel')
        data_write(file, self, 'drytolerance')
        data_write(file, self, 'wavetolerance')
        data_write(file, self, 'depthdeep')
        data_write(file, self, 'maxleveldeep')
        data_write(file, self, 'ifriction')
        data_write(file, self, 'coeffmanning')
        data_write(file, self, 'frictiondepth')
        file.close()

        print('Creating data file settopo.data')
        # open file and write a warning header:
        file = open_datafile('settopo.data')
        self.ntopofiles = len(self.topofiles)
        data_write(file, self, 'ntopofiles')
        for tfile in self.topofiles:
            try:
                fname = os.path.abspath(tfile[-1])
            except:
                print("*** Error: file not found: ",tfile[-1])
                raise MissingFile("file not found")
            file.write("\n'%s' \n " % fname)
            file.write("%3i %3i %3i %20.10e %20.10e \n" % tuple(tfile[:-1]))
        file.close()

        print('Creating data file setdtopo.data')
        # open file and write a warning header:
        file = open_datafile('setdtopo.data')
        self.mdtopofiles = len(self.dtopofiles)
        data_write(file, self, 'mdtopofiles')
        data_write(file, self, None)
        for tfile in self.dtopofiles:
            try:
                fname = "'%s'" % os.path.abspath(tfile[-1])
            except:
                print("*** Error: file not found: ",tfile[-1])
                raise MissingFile("file not found")
            file.write("\n%s \n" % fname)
            file.write("%3i %3i %3i\n" % tuple(tfile[:-1]))
        file.close()

        print('Creating data file setqinit.data')
        # open file and write a warning header:
        file = open_datafile('setqinit.data')
        # self.iqinit tells which component of q is perturbed!
        data_write(file, self, 'iqinit')
        data_write(file, self, None)
        for tfile in self.qinitfiles:
            try:
                fname = "'%s'" % os.path.abspath(tfile[-1])
            except:
                print("*** Error: file not found: ",tfile[-1])
                raise MissingFile("file not found")
            file.write("\n%s  \n" % fname)
            file.write("%3i %3i \n" % tuple(tfile[:-1]))
        file.close()

        make_setgauges_datafile(self)

#        print 'Creating data file setgauges.data'
#        # open file and write a warning header:
#        file = open_datafile('setgauges.data')
#        self.ngauges = len(self.gauges)
#        data_write(file, self, 'ngauges')
#        data_write(file, self, None)
#        gaugeno_used = []
#        for gauge in self.gauges:
#            gaugeno = gauge[0]
#            if gaugeno in gaugeno_used:
#                print "*** Gauge number %s used more than once! " % gaugeno
#                raise Exception("Repeated gauge number")
#            else:
#                gaugeno_used.append(gauge[0])
#            gauge.append(gaugeno)
#            file.write("%3i %19.10e  %19.10e  %15.6e  %15.6e  =: gauge%s\n" % tuple(gauge))
#        file.close()


        print('Creating data file setfixedgrids.data')
        # open file and write a warning header:
        file = open_datafile('setfixedgrids.data')
        self.nfixedgrids = len(self.fixedgrids)
        data_write(file, self, 'nfixedgrids')
        data_write(file, self, None)
        for fixedgrid in self.fixedgrids:
            file.write(11*"%g  " % tuple(fixedgrid) +"\n")
        file.close()


        print('Creating data file setregions.data')
        # open file and write a warning header:
        file = open_datafile('setregions.data')
        self.nregions = len(self.regions)
        data_write(file, self, 'nregions')
        data_write(file, self, None)
        for regions in self.regions:
            file.write(8*"%g  " % tuple(regions) +"\n")
        file.close()




r"""
Module containing SharpClaw solvers for PyClaw/PetClaw

#  File:        sharpclaw.py
#  Created:     2010-03-20
#  Author:      David Ketcheson
"""
# Solver superclass
from clawpack.pyclaw.solver import Solver
from clawpack.pyclaw.util import add_parent_doc

# Reconstructor
try:
    # load c-based WENO reconstructor (PyWENO)
    from clawpack.pyclaw.limiters import reconstruct as recon
except ImportError:
    # load old WENO5 reconstructor
    from clawpack.pyclaw.limiters import recon

def default_tfluct(self):
    r"""This is a dummy routine and should never be called, check Euler1D
        to learn how to pass tfluct functions to the sharpclaw solver
    """
    if self.tfluct_solver:
        raise Exception("You set solver.tfluct_solver=True, but solver.tfluct has not been set.")
    pass

class SharpClawSolver(Solver):
    r"""
    Superclass for all SharpClawND solvers.

    Implements Runge-Kutta time stepping and the basic form of a 
    semi-discrete step (the dq() function).  If another method-of-lines
    solver is implemented in the future, it should be based on this class,
    which then ought to be renamed to something like "MOLSolver".

    .. attribute:: lim_type

        Limiter(s) to be used.
            - 0: No limiting.
            - 1: TVD reconstruction.
            - 2: WENO reconstruction.

        ``Default = 2``

    .. attribute:: weno_order

        Order of the WENO reconstruction. From 1st to 17th order (PyWENO)

        ``Default = 5``

    .. attribute:: time_integrator

        Time integrator to be used. Currently implemented methods:

            - 'Euler'  : 1st-order Forward Euler integration
            - 'SSP33'  : 3rd-order strong stability preserving method of Shu & Osher
            - 'SSP104' : 4th-order strong stability preserving method Ketcheson
            - 'SSPLMM32': 2nd-order strong stability preserving 3-step linear multistep method,
                          using Euler for starting values
            - 'SSPLMM43': 3rd-order strong stability preserving 4-step linear multistep method
                          using SSPRK22 for starting values
            - 'RK'     : Arbitrary Runge-Kutta method, specified by setting `solver.a`
                         and `solver.b` to the Butcher arrays of the method.
            - 'LMM'    : Arbitrary linear multistep method, specified by setting the
                         coefficient arrays `solver.alpha` and `solver.beta`.

        ``Default = 'SSP104'``

    .. attribute:: char_decomp

        Type of WENO reconstruction.
        0: conservative variables WENO reconstruction (standard).
        1: Wave-slope reconstruction.
        2: characteristic-wise WENO reconstruction.
        3: transmission-based WENO reconstruction.
        ``Default = 0``

    .. attribute:: tfluct_solver

        Whether a total fluctuation solver have to be used. If True the function
        that calculates the total fluctuation must be provided.
        ``Default = False``

    .. attribute:: tfluct

        Pointer to Fortran routine to calculate total fluctuation
        ``Default = default_tfluct (None)``

    .. attribute:: aux_time_dep

        Whether the auxiliary array is time dependent.
        ``Default = False``
    
    .. attribute:: kernel_language

        Specifies whether to use wrapped Fortran routines ('Fortran')
        or pure Python ('Python').  
        ``Default = 'Fortran'``.

    .. attribute:: num_ghost

        Number of ghost cells.
        ``Default = 3``

    .. attribute:: fwave
    
        Whether to split the flux jump (rather than the jump in Q) into waves; 
        requires that the Riemann solver performs the splitting.  
        ``Default = False``

    .. attribute:: cfl_desired

        Desired CFL number.
        ``Default = 2.45``

    .. attribute:: cfl_max

        Maximum CFL number.
        ``Default = 2.50``

    .. attribute:: dq_src

        Whether a source term is present. If it is present the function that 
        computes its contribution must be provided.
        ``Default = None``

    .. attribute:: call_before_step_each_stage

        Whether to call the method `self.before_step` before each RK stage.
        ``Default = False``

    """
    _sspcoeff = {
       'Euler' :    1.0,
       'SSP22':     1.0,
       'SSP33':     1.0,
       'SSP104':    6.0,
       'SSPLMM32':  0.5,
       'SSPLMM43':  1./3.,
       'SSPLMM53':  0.5,
       'RK':        None,
       'LMM':       None
       }

    _cfl_default = {
        'SSP104':   [2.45, 2.5],
        'SSPLMM32': [0.24, 0.25],
        'SSPLMM43': [0.15, 1./6.],
        'SSPLMM53': [0.24, 0.25]
        }

    # ========================================================================
    #   Initialization routines
    # ========================================================================
    def __init__(self,riemann_solver=None,claw_package=None):
        r"""
        Set default options for SharpClawSolvers and call the super's __init__().
        """
        self.limiters = [1]
        self.lim_type = 2
        self.weno_order = 5
        self.time_integrator = 'SSP104'
        self.char_decomp = 0
        self.tfluct_solver = False
        self.tfluct = default_tfluct
        self.aux_time_dep = False
        self.kernel_language = 'Fortran'
        self.num_ghost = 3
        self.fwave = False
        self.cfl_desired = None
        self.cfl_max = None
        self.dq_src = None
        self.call_before_step_each_stage = False
        self._mthlim = self.limiters
        self._method = None
        self._registers = None
        self.dq_dt = None
        self.dt_old = None        

        # Used only if time integrator is 'RK'
        self.a = None
        self.b = None
        self.c = None

        # Used only if time integrator is a multistep method
        self.sspcoeff0 = None
        self.alpha = None
        self.beta = None
        self.lmm_steps = 4
        self.sspcoeff = None
        self.prev_dq_dt_values = []
        self.prev_dt_values = []
        self.prev_dtFE_values = []
        self.check_lmm_cond = False
        self.lmm_cond = True

        # Call general initialization function
        super(SharpClawSolver,self).__init__(riemann_solver,claw_package)

 
    def setup(self,solution):
        """
        Allocate RK stage arrays or previous step solutions and fortran routine work arrays.
        """
        if self.lim_type == 2:
            self.num_ghost = (self.weno_order+1)//2

        if self.lim_type == 2 and self.weno_order != 5 and self.kernel_language == 'Python':
            raise Exception('Only 5th-order WENO reconstruction is implemented in Python kernels. \
                             Use Fortran for higher-order WENO.')
        # This is a hack to deal with the fact that petsc4py
        # doesn't allow us to change the stencil_width (num_ghost)
        state = solution.state
        state.set_num_ghost(self.num_ghost)
        # End hack

        if self.time_integrator == 'LMM':
            assert self.dt_variable == False, \
                'Must set solver.dt_variable=False for LMM integrator.'
        try:
            if 'SSPLMM' in self.time_integrator:
                if self.time_integrator == 'SSPLMMk2':
                    assert 3 <= self.lmm_steps, \
                        'Must set solver.lmm_steps greater than 2 for 2nd order SSPLMM integrator.'
                    self.sspcoeff = (self.lmm_steps - 2.)/(self.lmm_steps - 1.)
                    # The choice of cfl_desired and cfl_max is intended for LMM with many steps (up to 20). 
                    # If more steps are chosen the solution may not be accurate enough.
                    # For a smaller number of steps, higher values of cfl_desired and cfl_max can be used.
                    if self.cfl_max is None:
                        self.cfl_desired = 0.14*self.sspcoeff 
                        self.cfl_max = 0.15*self.sspcoeff 
                elif self.time_integrator == 'SSPLMMk3':
                    assert 4 <= self.lmm_steps <= 5, \
                        'Must set solver.lmm_steps equal to 4 or 5 for 3rd order SSPLMM integrator.'
                    self.sspcoeff = (self.lmm_steps - 3.)/(self.lmm_steps - 1.)
                    if self.cfl_max is None:
                        self.cfl_desired = 0.48*self.sspcoeff 
                        self.cfl_max = 0.5*self.sspcoeff 
            else:
                if self.cfl_max is None:
                    self.cfl_desired = self._cfl_default[self.time_integrator][0]
                    self.cfl_max = self._cfl_default[self.time_integrator][1]
            if self.cfl_desired is None:
                self.cfl_desired = 0.9*self.cfl_max
        except KeyError:
            raise KeyError('Maximum CFL number is not provided.')

        self._allocate_registers(solution)
        self._set_mthlim()

        state = solution.states[0]
 
        if self.kernel_language=='Fortran':
            if self.fmod is None:
                so_name = 'clawpack.pyclaw.sharpclaw.sharpclaw'+str(self.num_dim)
                self.fmod = __import__(so_name,fromlist=['clawpack.pyclaw.sharpclaw'])
            state.set_cparam(self.fmod)
            state.set_cparam(self.rp)
            state.set_cparam(self.tfluct)
            self._set_fortran_parameters(state,self.fmod.clawparams,self.fmod.workspace,self.fmod.reconstruct)

        self._allocate_bc_arrays(state)

        super(SharpClawSolver,self).setup(solution)


    def __del__(self):
        r"""
        Deallocate F90 module arrays.
        Also delete Fortran objects, which otherwise tend to persist in Python sessions.
        """
        if self.kernel_language=='Fortran':
            self.fmod.clawparams.dealloc_clawparams()
            self.fmod.workspace.dealloc_workspace(self.char_decomp)
            self.fmod.reconstruct.dealloc_recon_workspace(self.fmod.clawparams.lim_type,self.fmod.clawparams.char_decomp)
            del self.fmod

        super(SharpClawSolver,self).__del__()


    # ========== Time stepping routines ======================================
    def step(self,solution,take_one_step,tstart,tend):
        """Evolve q over one time step.

        Take one step with a Runge-Kutta or multistep method as specified by
        `solver.time_integrator`.
        """
        state = solution.states[0]
        step_index = self.status['numsteps'] + 1
        if self.accept_step == True:
            self.cfl.set_global_max(0.)
            self.dq_dt = self.dq(state) / self.dt

        if 'LMM' in self.time_integrator:
            step_index = self.update_saved_values(state,step_index)

        self.get_dt(solution.t,tstart,tend,take_one_step)

        # Recompute cfl number based on current step-size
        cfl = self.cfl.get_cached_max()
        self.cfl.set_global_max(self.dt / self.dt_old * cfl)
        self.dt_old = self.dt

        ### Runge-Kutta methods ###
        if self.time_integrator == 'Euler':
            state.q += self.dt*self.dq_dt

        elif self.time_integrator == 'SSP22':
            self.ssp22(state)

        elif self.time_integrator == 'SSP33':
            self._registers[0].q = state.q + self.dt*self.dq_dt
            self._registers[0].t = state.t + self.dt

            if self.call_before_step_each_stage:
                self.before_step(self,self._registers[0])
            self._registers[0].q = 0.75*state.q + 0.25*(self._registers[0].q + self.dq(self._registers[0]))
            self._registers[0].t = state.t + 0.5*self.dt

            if self.call_before_step_each_stage:
                self.before_step(self,self._registers[0])
            state.q = 1./3.*state.q + 2./3.*(self._registers[0].q + self.dq(self._registers[0]))

        elif self.time_integrator == 'SSP104':
            self.ssp104(state)

        elif self.time_integrator == 'RK':
            # General explicit RK with specified coefficients

            # This is pulled out of the loop in order to use dq_dt
            self._registers[0].q = self.dt*self.dq_dt
            self._registers[0].t = state.t

            num_stages = len(self.b)
            for i in range(1,num_stages):
                self._registers[i].q[:] = state.q
                self._registers[i].t = state.t + self.dt*self.c[i]
                if self.call_before_step_each_stage:
                    self.before_step(self,self._registers[i])
                for j in range(i):
                    self._registers[i].q += self.a[i,j]*self._registers[j].q

                # self._registers[i].q eventually stores dt*f(y_i) after stage solution y_i is computed
                self._registers[i].q = self.dq(self._registers[i])

            for j in range(num_stages):
                state.q += self.b[j]*self._registers[j].q

        ### Linear multistep methods ###
        elif self.time_integrator in ['SSPLMMk2', 'SSPLMMk3']:
            num_steps = self.lmm_steps
            if step_index < num_steps:
                # Use SSP22 Runge-Kutta method for starting values
                self.ssp22(state)
            else:
                if self.time_integrator == 'SSPLMMk2':
                    omega_k_minus_1 = sum(self.prev_dt_values[1:])/self.dt
                    r = (omega_k_minus_1-1.)/omega_k_minus_1 # SSP coefficient

                    delta = 1./omega_k_minus_1**2
                    beta = (omega_k_minus_1+1.)/omega_k_minus_1
                    state.q = beta*(r*state.q + self.dt*self.dq_dt) + delta*self._registers[-num_steps].q
                else:
                    omega_k_minus_1 = sum(self.prev_dt_values[1:])/self.dt
                    omega_k = omega_k_minus_1 + 1.
                    r = (omega_k_minus_1-2.)/omega_k_minus_1 # SSP coefficient

                    delta0 = (4*omega_k - omega_k_minus_1**2)/omega_k_minus_1**3
                    beta0 = omega_k/omega_k_minus_1**2
                    beta_k_minus_1 = omega_k**2/omega_k_minus_1**2

                    state.q = beta_k_minus_1*(r*state.q + self.dt*self.prev_dq_dt_values[-1]) + \
                            (r*beta0 + delta0)*self._registers[-num_steps].q + \
                            beta0*self.dt*self.prev_dq_dt_values[-num_steps]

        elif self.time_integrator == 'LMM':
            if step_index < len(self._registers):
                self.ssp104(state) # Use SSP104 for starting values
            else:
                # Update solution: alpha[-1] and beta[-1] correspond to solution at the previous step
                state.q = self.alpha[-1]*self._registers[-1].q + self.beta[-1]*self.dt*self.prev_dq_dt_values[-1]
                for i in range(self.lmm_steps-1):
                    state.q += self.alpha[i]*self._registers[i].q + self.beta[i]*self.dt*self.prev_dq_dt_values[i]

        else:
            raise Exception('Unrecognized time integrator')
            return False


    def ssp22(self,state):
        self._registers[0].q = state.q + self.dt*self.dq_dt
        self._registers[0].t = state.t + self.dt

        if self.call_before_step_each_stage:
            self.before_step(self,self._registers[0])
        state.q = 0.5*(state.q + self._registers[0].q + self.dq(self._registers[0]))


    def ssp104(self,state):
        if self.time_integrator == 'SSP104':
            s1 = self._registers[0]
            s1.q[:] = state.q
        elif self.time_integrator == 'LMM':
            # Okay to copy state objects here since this only happens a few times
            import copy
            s1 = copy.deepcopy(state)

        s1.q = state.q + self.dt*self.dq_dt/6.
        s1.t = state.t + self.dt/6.

        for i in range(4):
            if self.call_before_step_each_stage:
                self.before_step(self,s1)
            s1.q = s1.q + self.dq(s1)/6.
            s1.t = s1.t + self.dt/6.

        state.q = state.q/25. + 0.36 * s1.q
        s1.q = 15. * state.q - 5. * s1.q
        s1.t = state.t + self.dt/3.

        for i in range(4):
            if self.call_before_step_each_stage:
                self.before_step(self,s1)
            s1.q = s1.q + self.dq(s1)/6.
            s1.t = s1.t + self.dt/6.

        if self.call_before_step_each_stage:
            self.before_step(self,s1)
        state.q += 0.6 * s1.q + 0.1 * self.dq(s1)


    def update_saved_values(self,state,step_index):
        r""" 
        Updates lists of saved function evaluations, solution values, dt and dtFE for LMMs.
        For 3rd-order SSPLMM additional conditions are checked if self.check_lmm_cond is set to True.
        If these conditions are violated, the step is rejected.
        """
        if (self.prev_dt_values == []):
            # This only happens at the very beginning of the computation
            self._registers[-1].q[:] = state.q
            self._registers[-1].t = state.t
            self.prev_dq_dt_values.append(self.dq_dt)
            self.prev_dt_values.append(0.) # Not used
            if 'SSPLMM' in self.time_integrator:
                cfl = self.cfl.get_cached_max()
                dtFE = self.dt / cfl * self.cfl_max / self.sspcoeff
                self.prev_dtFE_values.append(dtFE)
                self.sspcoeff0 = self._sspcoeff['SSP22']
        elif self.accept_step == True: # Previous step was accepted
            if 'SSPLMM' in self.time_integrator:
                cfl = self.cfl.get_cached_max()
                dtFE = self.dt / cfl * self.cfl_max / self.sspcoeff

                if self.time_integrator == 'SSPLMMk3' and self.check_lmm_cond:
                    self.lmm_cond = self.check_3rd_ord_cond(state,step_index,dtFE)
                    if not self.lmm_cond:
                        self.accept_step = False
                        state.q[:] = self._registers[-1].q
                        self.dq_dt = self.prev_dq_dt_values[-1]
                        state.t = self._registers[-1].t
                        self.status['numsteps'] -= 1
                        return self.status['numsteps'] + 1

                if step_index <= len(self._registers):  # Startup
                    if self.time_integrator == 'SSPLMMk3':
                        self.prev_dq_dt_values.append(self.dq_dt)
                    self.prev_dt_values.append(self.dt_old)
                    self.prev_dtFE_values.append(dtFE)
                else:
                    if self.time_integrator == 'SSPLMMk3':
                        self.prev_dq_dt_values = self.prev_dq_dt_values[1:] + self.prev_dq_dt_values[:1]
                        self.prev_dq_dt_values[-1] = self.dq_dt

                    # Roll and update prev_dt_values and prev_dtFE_values lists
                    self.prev_dt_values = self.prev_dt_values[1:] + self.prev_dt_values[:1]
                    self.prev_dt_values[-1] = self.dt_old
                    self.prev_dtFE_values = self.prev_dtFE_values[1:] + self.prev_dtFE_values[:1]
                    self.prev_dtFE_values[-1] = dtFE
            elif self.time_integrator == 'LMM':
                if step_index <= len(self._registers):
                    self.prev_dq_dt_values.append(self.dq_dt)
                else:
                    self.prev_dq_dt_values = self.prev_dq_dt_values[1:] + self.prev_dq_dt_values[:1]
                    self.prev_dq_dt_values[-1] = self.dq_dt

            # Roll and update saved solution
            self._registers = self._registers[1:] + self._registers[:1]
            self._registers[-1].q[:] = state.q
            self._registers[-1].t = state.t

        return step_index


    def check_3rd_ord_cond(self,state,step_index,dtFE):
        r"""
        This routine checks the additional conditions for the 3rd-order SSPLMMs.
        This is a posteriori check after a step is accepted.
        In particular, there is a condition on the step size for the starting values and 
        a condition on the ratio of forward Euler step sizes at very step.
        If the conditions are violated we muct retrieve the previous solution and discard
        that step; otherwise the step is accepted.
        """
        lmm_cond = True

        if step_index <= len(self._registers):
            rho = 0.6 if len(self._registers) == 4 else 0.57
            if self.dt > rho * dtFE:
                lmm_cond = False

        rhoFE = 0.9 if len(self._registers) == 4 else 0.962
        dtFEm1 = self.prev_dtFE_values[-1]
        if rhoFE * dtFEm1 > dtFE or dtFE > dtFEm1 / rhoFE:
            lmm_cond = False

        return lmm_cond


    def _set_mthlim(self):
        self._mthlim = self.limiters
        if not isinstance(self.limiters,list): self._mthlim=[self._mthlim]
        if len(self._mthlim)==1: self._mthlim = self._mthlim * self.num_waves
        if len(self._mthlim)!=self.num_waves:
            raise Exception('Length of solver.limiters is not equal to 1 or to solver.num_waves')

       
    def dq(self,state):
        """
        Evaluate dq/dt * (delta t)
        """

        deltaq = self.dq_hyperbolic(state)

        if self.dq_src is not None:
            deltaq+=self.dq_src(self,state,self.dt)

        return deltaq


    def dq_hyperbolic(self,state):
        raise NotImplementedError('You must subclass SharpClawSolver.')

         
    def dqdt(self,state):
        """
        Evaluate dq/dt.  This routine is used for implicit time stepping.
        """

        self.dt = 1
        dq_dt = self.dq_hyperbolic(state)

        if self.dq_src is not None:
            dq_dt += self.dq_src(self,state,self.dt)

        return dq_dt.flatten('f')


    def _set_fortran_parameters(self,state,clawparams,workspace,reconstruct):
        """
        Set parameters for Fortran modules used by SharpClaw.
        The modules should be imported and passed as arguments to this function.
        """
        grid = state.grid
        clawparams.num_dim       = grid.num_dim
        clawparams.lim_type      = self.lim_type
        clawparams.weno_order    = self.weno_order
        clawparams.char_decomp   = self.char_decomp
        clawparams.tfluct_solver = self.tfluct_solver
        clawparams.fwave         = self.fwave
        clawparams.index_capa         = state.index_capa+1

        clawparams.num_waves     = self.num_waves
        clawparams.alloc_clawparams()
        for idim in range(grid.num_dim):
            clawparams.xlower[idim]=grid.dimensions[idim].lower
            clawparams.xupper[idim]=grid.dimensions[idim].upper
        clawparams.dx       =grid.delta
        clawparams.mthlim   =self._mthlim

        maxnx = max(grid.num_cells)+2*self.num_ghost
        workspace.alloc_workspace(maxnx,self.num_ghost,state.num_eqn,self.num_waves,self.char_decomp)
        reconstruct.alloc_recon_workspace(maxnx,self.num_ghost,state.num_eqn,self.num_waves,
                                            clawparams.lim_type,clawparams.char_decomp)


    def _allocate_registers(self,solution):
        r"""
        Instantiate State objects for Runge--Kutta stages and Linear Multistep method steps.

        This routine is only used by method-of-lines solvers (SharpClaw),
        not by the Classic solvers.  It allocates additional State objects
        to store the intermediate stages used by Runge--Kutta and Multistep 
        time integrators.

        If we create a MethodOfLinesSolver subclass, this should be moved there.
        """
        # Generally the number of registers for the starting method should be at most 
        # equal to the number of registers of the LMM
        if self.time_integrator   == 'Euler':   nregisters=0
        elif self.time_integrator == 'SSP33':   nregisters=1
        elif self.time_integrator == 'SSP104':  nregisters=1
        elif self.time_integrator == 'RK':      nregisters=len(self.b)+1
        elif self.time_integrator == 'SSPLMMk2': nregisters=self.lmm_steps
        elif self.time_integrator == 'SSPLMMk3': nregisters=self.lmm_steps
        elif self.time_integrator == 'LMM': nregisters=len(self.alpha)
        else:
            raise Exception('Unrecognized time integrator: '+self.time_integrator)
        
        state = solution.states[0]
        # Use the same class constructor as the solution for the Runge Kutta stages
        self._registers = []
        for i in range(nregisters):
            import copy
            self._registers.append(copy.deepcopy(state))


    def accept_reject_step(self,state):
        r"""
        Decide whether to accept or not the current step.
        For Runge-Kutta methods the step is accepted if cfl <= cfl_max.
        For SSPLMM32 the choice of step-size guarantees the cfl condition is satisfied for the steps the LMM
        is used. Hence, we need to check the cfl condition only for the starting steps. 
        """
        accept_step = True
        cfl = self.cfl.get_cached_max()

        if 'LMM' in self.time_integrator:
            step_index = self.status['numsteps'] + 1

            # Condition for starting RK methods
            if step_index < len(self._registers):
                if self.time_integrator == 'LMM':
                    sspcoeff_ratio = 1.
                else:
                    sspcoeff_ratio = self.sspcoeff0/self.sspcoeff

                if cfl > sspcoeff_ratio * self.cfl_max:
                    accept_step = False

        # Check cfl condition for Runge-Kutta methods
        else:
            if cfl > self.cfl_max:
                accept_step = False

        return accept_step


    def get_dt_new(self):
        r"""
        Set size of next step depending on the time integrator and
        whether or not the current step was accepted.
        """
        self.dt_old = self.dt        
        cfl = self.cfl.get_cached_max()

        if 'SSPLMM' in self.time_integrator:
            step_index = self.status['numsteps'] + 1

            if step_index < len(self._registers):
                # Step-size update of starting methods
                sspcoeff_ratio = self.sspcoeff0/self.sspcoeff
                self.dt = sspcoeff_ratio * self.dt * self.cfl_desired / cfl
                if self.time_integrator == 'SSPLMMk3' and self.check_lmm_cond and not self.lmm_cond:
                    rho = 0.6 if len(self._registers)== 4 else 0.57
                    self.dt = rho * self.dt
            else:
                # Step size selection guarantees CFL condition is satisfied.
                # Only need to check 3rd-order LMM's condition
                if self.accept_step:
                    s = len(self._registers)
                    p = int(self.time_integrator[-1])
                    mu = min([self.prev_dtFE_values[i] for i in range(s)])
                    H = sum(self.prev_dt_values[1:])
                    self.dt = H * mu / (H + (p-1)*mu)
                elif self.time_integrator == 'SSPLMMk3' and self.check_lmm_cond:
                    self.dt = 0.5 * self.dt

        # Step-size update for RK methods
        else:
            self.dt = self.dt * self.cfl_desired / cfl



# ========================================================================
class SharpClawSolver1D(SharpClawSolver):
# ========================================================================
    """
    SharpClaw solver for one-dimensional problems.
    
    Used to solve 1D hyperbolic systems using the SharpClaw algorithms,
    which are based on WENO reconstruction and Runge-Kutta time stepping.
    """

    __doc__ += add_parent_doc(SharpClawSolver)
    
    def __init__(self,riemann_solver=None,claw_package=None):
        r"""
        See :class:`SharpClawSolver1D` for more info.
        """   
        self.num_dim = 1
        self.reflect_index = [1]
        super(SharpClawSolver1D,self).__init__(riemann_solver,claw_package)


    def dq_hyperbolic(self,state):
        r"""
        Compute dq/dt * (delta t) for the hyperbolic hyperbolic system.

        Note that the capa array, if present, should be located in the aux
        variable.

        Indexing works like this (here num_ghost=2 as an example)::

         0     1     2     3     4     mx+num_ghost-2     mx+num_ghost      mx+num_ghost+2
                     |                        mx+num_ghost-1 |  mx+num_ghost+1
         |     |     |     |     |   ...   |     |     |     |     |
            0     1  |  2     3            mx+num_ghost-2    |mx+num_ghost       
                                                  mx+num_ghost-1   mx+num_ghost+1

        The top indices represent the values that are located on the grid
        cell boundaries such as waves, s and other Riemann problem values, 
        the bottom for the cell centered values such as q.  In particular
        the ith grid cell boundary has the following related information::

                          i-1         i         i+1
                           |          |          |
                           |   i-1    |     i    |
                           |          |          |

        Again, grid cell boundary quantities are at the top, cell centered
        values are in the cell.

        """
    
        import numpy as np

        self._apply_bcs(state)
        q = self.qbc
        aux = self.auxbc 

        grid = state.grid
        mx = grid.num_cells[0]

        ixy=1

        if self.kernel_language=='Fortran':
            rp1 = self.rp.rp1._cpointer
            if self.tfluct_solver:
                tfluct1 = self.tfluct.tfluct1._cpointer
            else:
                tfluct1 = self.tfluct

            dq,cfl=self.fmod.flux1(q,self.auxbc,self.dt,state.t,ixy,mx,self.num_ghost,mx,rp1,tfluct1)

        elif self.kernel_language=='Python':

            dtdx = np.zeros( (mx+2*self.num_ghost) ,order='F')
            dq   = np.zeros( (state.num_eqn,mx+2*self.num_ghost) ,order='F')

            # Find local value for dt/dx
            if state.index_capa>=0:
                dtdx = self.dt / (grid.delta[0] * aux[state.index_capa,:])
            else:
                dtdx += self.dt/grid.delta[0]
 
            if aux.shape[0]>0:
                aux_l=aux[:,:-1]
                aux_r=aux[:,1: ]
            else:
                aux_l = None
                aux_r = None

            #Reconstruct (wave reconstruction uses a Riemann solve)
            if self.lim_type==-1: #1st-order Godunov
                ql=q; qr=q
            elif self.lim_type==0: #Unlimited reconstruction
                raise NotImplementedError('Unlimited reconstruction not implemented')
            elif self.lim_type==1: #TVD Reconstruction
                raise NotImplementedError('TVD reconstruction not implemented')
            elif self.lim_type==2: #WENO Reconstruction
                if self.char_decomp==0: #No characteristic decomposition
                    ql,qr=recon.weno(5,q)
                elif self.char_decomp==1: #Wave-based reconstruction
                    q_l=q[:,:-1]
                    q_r=q[:,1: ]
                    wave,s,amdq,apdq = self.rp(q_l,q_r,aux_l,aux_r,state.problem_data)
                    ql,qr=recon.weno5_wave(q,wave,s)
                elif self.char_decomp==2: #Characteristic-wise reconstruction
                    raise NotImplementedError

            # Solve Riemann problem at each interface
            q_l=qr[:,:-1]
            q_r=ql[:,1: ]
            wave,s,amdq,apdq = self.rp(q_l,q_r,aux_l,aux_r,state.problem_data)

            # Loop limits for local portion of grid
            # THIS WON'T WORK IN PARALLEL!
            LL = self.num_ghost - 1
            UL = grid.num_cells[0] + self.num_ghost + 1

            # Compute maximum wave speed
            cfl = 0.0
            for mw in range(self.num_waves):
                smax1 = np.max( dtdx[LL  :UL]  *s[mw,LL-1:UL-1])
                smax2 = np.max(-dtdx[LL-1:UL-1]*s[mw,LL-1:UL-1])
                cfl = max(cfl,smax1,smax2)

            #Find total fluctuation within each cell
            wave,s,amdq2,apdq2 = self.rp(ql,qr,aux,aux,state.problem_data)

            # Compute dq
            for m in range(state.num_eqn):
                dq[m,LL:UL] = -dtdx[LL:UL]*(amdq[m,LL:UL] + apdq[m,LL-1:UL-1] \
                                + apdq2[m,LL:UL] + amdq2[m,LL:UL])

        else: 
            raise Exception('Unrecognized value of solver.kernel_language.')

        self.cfl.update_global_max(cfl)
        return dq[:,self.num_ghost:-self.num_ghost]
    

# ========================================================================
class SharpClawSolver2D(SharpClawSolver):
# ========================================================================
    """ Two Dimensional SharpClawSolver
    """

    __doc__ += add_parent_doc(SharpClawSolver)

    def __init__(self,riemann_solver=None,claw_package=None):
        r"""
        Create 2D SharpClaw solver
        
        See :class:`SharpClawSolver2D` for more info.
        """   
        self.num_dim = 2
        self.reflect_index = [1,2]

        super(SharpClawSolver2D,self).__init__(riemann_solver,claw_package)


    def dq_hyperbolic(self,state):
        """Compute dq/dt * (delta t) for the hyperbolic hyperbolic system

        Note that the capa array, if present, should be located in the aux
        variable.

        Indexing works like this (here num_ghost=2 as an example)::

         0     1     2     3     4     mx+num_ghost-2     mx+num_ghost      mx+num_ghost+2
                     |                        mx+num_ghost-1 |  mx+num_ghost+1
         |     |     |     |     |   ...   |     |     |     |     |
            0     1  |  2     3            mx+num_ghost-2    |mx+num_ghost       
                                                  mx+num_ghost-1   mx+num_ghost+1

        The top indices represent the values that are located on the grid
        cell boundaries such as waves, s and other Riemann problem values, 
        the bottom for the cell centered values such as q.  In particular
        the ith grid cell boundary has the following related information::

                          i-1         i         i+1
                           |          |          |
                           |   i-1    |     i    |
                           |          |          |

        Again, grid cell boundary quantities are at the top, cell centered
        values are in the cell.

        """
        self._apply_bcs(state)
        q = self.qbc 

        grid = state.grid

        num_ghost=self.num_ghost
        mx=grid.num_cells[0]
        my=grid.num_cells[1]
        maxm = max(mx,my)

        if self.kernel_language=='Fortran':
            rpn2 = self.rp.rpn2._cpointer
            if self.tfluct_solver:
                tfluct2 = self.tfluct.tfluct2._cpointer
            else:
                tfluct2 = self.tfluct

            dq,cfl=self.fmod.flux2(q,self.auxbc,self.dt,state.t,num_ghost,maxm,mx,my,rpn2,tfluct2)

        else: raise Exception('Only Fortran kernels are supported in 2D.')

        self.cfl.update_global_max(cfl)
        return dq[:,num_ghost:-num_ghost,num_ghost:-num_ghost]

# ========================================================================
class SharpClawSolver3D(SharpClawSolver):
# ========================================================================
    """ Three Dimensional SharpClawSolver
    """

    __doc__ += add_parent_doc(SharpClawSolver)

    def __init__(self,riemann_solver=None,claw_package=None):
        r"""
        Create 3D SharpClaw solver
        
        See :class:`SharpClawSolver3D` for more info.
        """   
        self.num_dim = 3
        self.reflect_index = [1,2,3]

        super(SharpClawSolver3D,self).__init__(riemann_solver,claw_package)


    def teardown(self):
        r"""
        Deallocate F90 module arrays.
        Also delete Fortran objects, which otherwise tend to persist in Python sessions.
        """
        if self.kernel_language=='Fortran':
            self.fmod.workspace.dealloc_workspace(self.char_decomp)
            self.fmod.reconstruct.dealloc_recon_workspace(self.fmod.clawparams.lim_type,self.fmod.clawparams.char_decomp)
            self.fmod.clawparams.dealloc_clawparams()
            del self.fmod


    def dq_hyperbolic(self,state):
        """Compute dq/dt * (delta t) for the hyperbolic hyperbolic system

        Note that the capa array, if present, should be located in the aux
        variable.

        Indexing works like this (here num_ghost=2 as an example)::

         0     1     2     3     4     mx+num_ghost-2     mx+num_ghost      mx+num_ghost+2
                     |                        mx+num_ghost-1 |  mx+num_ghost+1
         |     |     |     |     |   ...   |     |     |     |     |
            0     1  |  2     3            mx+num_ghost-2    |mx+num_ghost       
                                                  mx+num_ghost-1   mx+num_ghost+1

        The top indices represent the values that are located on the grid
        cell boundaries such as waves, s and other Riemann problem values, 
        the bottom for the cell centered values such as q.  In particular
        the ith grid cell boundary has the following related information::

                          i-1         i         i+1
                           |          |          |
                           |   i-1    |     i    |
                           |          |          |

        Again, grid cell boundary quantities are at the top, cell centered
        values are in the cell.

        """
        self._apply_bcs(state)
        q = self.qbc 

        grid = state.grid

        num_ghost=self.num_ghost
        mx=grid.num_cells[0]
        my=grid.num_cells[1]
        mz=grid.num_cells[2]
        maxm = max(mx,my,mz)

        if self.kernel_language=='Fortran':
            rpn3 = self.rp.rpn3._cpointer
            if self.tfluct_solver:            
                tfluct3 = self.tfluct.tfluct3._cpointer
            else:
                tfluct3 = self.tfluct

            dq,cfl=self.fmod.flux3(q,self.auxbc,self.dt,state.t,num_ghost,maxm,mx,my,mz,rpn3,tfluct3)

        else: raise Exception('Only Fortran kernels are supported in 3D.')

        self.cfl.update_global_max(cfl)
        return dq[:,num_ghost:-num_ghost,num_ghost:-num_ghost,num_ghost:-num_ghost]



"""
Module for converting MOST data and output to GeoClaw format.

This is Work In Progress -- this module should be improved to use newer
topotools.

"""

import os, glob, re

def most2tt3(fname, fname2=None):
    """
    Converts MOST topo file to tt3 format.
    """
    f = open(fname).readlines()
    mn = f[0].split()
    ncols = int(mn[0])
    nrows = int(mn[1])
    xll = float(f[1]) 
    dx = float(f[2]) - xll
    xll = xll - 360.
    yll = float(f[nrows+ncols])
    dy =  float(f[nrows+ncols-1]) - yll
    if abs(dx-dy) > 1.e-6:
        print('*** WARNING: dx = ',dx,'  dy = ',dy)
    cellsize = dx

    if fname2 is None:
        fname2 = os.path.splitext(fname)[0] + '.asc'

    f2 = open(fname2,'w')
    f2.write('%s ncols\n%s nrows\n%s xll\n%s yll\n%s cellsize\n99999 nodata_value\n' \
          % (ncols,nrows,xll,yll,cellsize))
    f2.writelines(f[nrows+ncols+1:])
    f2.close()
    print("Created ",fname2)

def most2fortt(fnameprefix):
    """
    Converts MOST output files to fort.t files.
    """
    files = glob.glob(r'%s*' % fnameprefix)
    files.sort()
    s = r"%s(?P<hours>[0-9]*)h(?P<minutes>[0-9]*)m(?P<seconds>[0-9]*)s" \
          % fnameprefix
    regexp = re.compile(s)
    frameno = 1
    for fname in files:
        result = regexp.search(fname)
        try:
            hours = result.group("hours")
            minutes = result.group("minutes")
            seconds = result.group("seconds")
        except:
            print("*** Cannot parse fname: ",fname)
            raise
        t = int(hours)*3600. + int(minutes)*60. + int(seconds)
        fortname = "fort.t" + str(frameno).zfill(4)
        f = open(fortname, 'w')
        f.write("%18.8e     time\n" % t)
        f.write("%5i                  meqn\n" % 1)
        f.write("%5i                  ngrids\n" % 1)
        f.write("%5i                  ndim\n" % 0)
        f.write("%5i                  maux\n" % 2)
        f.close()
        print("Created %s from %s at time t = %s" % (fortname, fname, t))
        frameno = frameno + 1


def most2fortq(fnameprefix):
    """
    Converts MOST output files to fort.q files.
    """
    files = glob.glob(r'%s*' % fnameprefix)
    files.sort()
    frameno = 1
    for fname in files:
        f = open(fname).readlines()
        mn = f[0].split()
        ncols = int(mn[0])
        nrows = int(mn[1])
        xll = float(f[1]) 
        dx = float(f[2]) - xll
        xll = xll - 360.
        yll = float(f[nrows+ncols])
        dy =  float(f[nrows+ncols-1]) - yll
        if abs(dx-dy) > 1.e-6:
            print('*** WARNING: dx = ',dx,'  dy = ',dy)
        cellsize = dx
    
        fortname = 'fort.q' + str(frameno).zfill(4)
        f2 = open(fortname,'w')
        f2.write("%5i                  grid_number\n" % 1)
        f2.write("%5i                  AMR_level\n" % 1)
        f2.write("%5i                  mx\n" % ncols)
        f2.write("%5i                  my\n" % nrows)
        f2.write("%5i                  xlow\n" % xll)
        f2.write("%5i                  ylow\n" % yll)
        f2.write("%5i                  dx\n" % dx)
        f2.write("%5i                  dy\n" % dy)
        f2.write("\n")

        for k in range(len(f)-1, nrows+ncols, -1):
            for s in f[k].split():
                z = float(s)
                f2.write("%18.8e\n" % z)
        f2.close()
        print("Created %s from %s" % (fortname,fname))
        frameno += 1

def get_comMIT_topo(x1,x2,y1,y2, dx='1m', fname='most_topo.most'):
    r"""
    Get a topo file from the comMIT Bathymetry server, see
      http://sift.pmel.noaa.gov/ComMIT/bathymetry_help.html
    Inputs:
      - *x1, x2* (floats) extent of longitude (between 0 and 360)
      - *y1, y2* (floats) extent of latitude (between -90 and 90))
      - *dx* (float or str) cellsize in decimal degrees, 
             or string such as '1m', '30s' for arcminutes or arcseconds.
        Note that etopo1 bathymetry is used to construct this file and then
        linear interpolation is used if the resolution requested is finer
        than 1 minute.  But onshore topography comes from a different database
        and may be finer.
      - *fname* (str) File name to save as.

    The result is in MOST format.  
    You can convert to GeoClaw form using most2tt3, e.g.
        >>> most2geoclaw.most2tt3('most_topo.most')
    But note that this seems to create a file with topo_type == -3.

    Retrieving seems to only work for some values of x1,x2,y1,y2 and not others
    for reasons not yet understood!
      
    """
    import os, time
    import subprocess
    fname_most = os.path.splitext(fname)[0] + '.most'

    s = "wget -O " + fname_most + \
        " 'http://sift.pmel.noaa.gov/ComMIT/bathymetry/create.most" + \
        "?xmin=%s&xmax=%s&ymin=%s&ymax=%s&cellsize=%s'"  % (x1,x2,y1,y2,dx)

    wait_time = 3
    time.sleep(wait_time)
    subprocess.check_call(s,shell=True)
    lines = open(fname_most).readlines()
    attempts = 1
    while (len(lines) < 2) and (attempts < 5):
        print("Status: ",lines[0])
        print("Will retry in %s seconds..." % wait_time)
        time.sleep(wait_time)
        subprocess.check_call(s,shell=True)
        lines = open(fname_most).readlines()
        attempts += 1
    if attempts<5:
        print("File downloaded with dimensions: ",lines[0])
    else:
        print("Tried %s times and file not ready... " % attempts)
        print("     wait a bit and try same call again")



if __name__=='__main__':
    import sys
    most2tt3(sys.argv[1])



"""
Print out 
  - settings of environment variables,
  - elements of sys.path that contain "clawpack"
  - lines of site-packages/easy-install.pth files that contain "clawpack"

Useful for figuring out how paths are set and which version of clawpack is being
used when more than one has been installed.

Notes: 
  - this should be run in the users' shell (which might not be bash)
  - this script has not been extensively tested, maybe it's missing something

"""


import sys, os
import site
import inspect

try:
    import clawpack
    claw_file = inspect.getfile(clawpack)
    claw_dir = os.path.split(claw_file)[0]
    print('\n`import clawpack` imports from:\n    %s' % claw_dir[:-9])
except:
    print('\n`import clawpack` gives an import error')
    

CLAW = os.environ.get('CLAW','')
if CLAW != '':
    print("\nThe CLAW environment variable is set to: \n    %s" % CLAW)
else:
    print("\nThe CLAW environment variable is not set")

PPATH = os.environ.get('PYTHONPATH','')
if PPATH != '':
    print("The PYTHONPATH environment variable is set to: \n    %s" % PPATH)
else:
    print("The PYTHONPATH environment variable is not set")

print("\nThe following directories on sys.path might contain clawpack,")
print("and are searched in this order:")
ppath = sys.path
for p in ppath:
    if 'clawpack' in p:
        if os.path.isdir(os.path.join(p,'clawpack')):
            print('    %s' % p)
        else:
            continue

print("\nThe following easy-install.pth files list clawpack:")
for p in ppath:
    if 'site-packages' in p:
        if os.path.isfile(os.path.join(p, 'easy-install.pth')):
            fname = os.path.join(p, 'easy-install.pth')
            f = open(fname).readlines()
            for line in f:
                if 'clawpack' in line:
                    print('    %s \n        (points to %s)' \
                          % (fname,line.strip()))
        else:
            continue







#!/usr/bin/env python
# coding: utf-8

# # Make Input Files 
# 
# ### For `$CLAW/geoclaw/examples/tsunami/eta_init_force_dry`
# 
# For this example simple artificial topography is generated in order to illustrate various things.
# 
# Contents:
# 
#  - [Define ocean topography](#topo_ocean)
#  - [Define topo for small coastal region](#topo_coast)
#  - [Create dtopo for an earthquake source](#dtopo)
#  - [Force Dry array](#force_dry)
#  
# Running this notebook should create a set of files in the directory `input_files`.
# 
# Alternatively, running 
# 
#     make input
#     
# or equivalently 
# 
#     python make_input_files.py
#     
# will run the python script version of this notebook, which was created with the command
# 
#     jupyter nbconvert --to python --TagRemovePreprocessor.enabled=True \
#             --TagRemovePreprocessor.remove_cell_tags="['hide-py']" \
#             make_input_files.ipynb
# 
# This will only work if [nbconvert](https://nbconvert.readthedocs.io/en/latest/index.html) is installed.
# 
# Note that cells in this notebook that create plots are not included in the `.py` version (due to the cell tag `hide-py` that is applied to these cells, visible if you select `View -> Cell Toolbar -> Tags` in the notebook menu).

# In[2]:


from pylab import *
from scipy.interpolate import interp1d
import os


# In[3]:


from clawpack.geoclaw import topotools, marching_front, dtopotools
from clawpack.visclaw import plottools


# ## Directory for input files:

# In[4]:


inputdir = 'input_files'
os.system('mkdir -p %s' % inputdir)
print('Input files will be put in directory %s' % inputdir)


# <div id="topo_ocean"></div>
# 
# ## Define ocean topography
# 
# This simple topography is piecewise linear in $x$ (longitude) with a continental shelf and beach, and constant in the $y$ (latitude) direction.  It is placed at the equator so distances are roughly equal in $x$ and $y$, and also placed at longitude 0.  

# In[5]:


# Define piecewise linear function (unequally spaced):
xocean = array([-2,-1,-0.5,-0.1,0.1])
zocean = array([-3000,-3000,-100,-100,100])

# Interpolate to equally spaced grid for topofile:
xo = arange(-2,0.2,0.1)
yo = array([-2,2])
zfunc = interp1d(xocean,zocean,fill_value="extrapolate")
zo = zfunc(xo)

# Convert to 2d arrays:
Xo,Yo = meshgrid(xo,yo)
Zo = vstack((zo,zo))


# ### Save as a topofile:

# In[7]:


topo = topotools.Topography()
topo.set_xyZ(xo,yo,Zo)

topofile = '%s/topo_ocean.tt3' % inputdir
topo.write(topofile, topo_type=3, Z_format="%11.3e")
print('Created ', topofile)


# <div id="topo_coast"></div>
# 
# ## Define topo for small coastal region
# 
# We define some more complicated topography on a finer grid over a small coastal region with 1/3 arcsecond resolution, chosen to be aligned with integer multiples of degrees (e.g. a grid point at longitude `x=0` and latitude `y=0`) as typical of real DEMs from NCEI.  This is important when aligning computational grids and fgmax grids (if used) in `setrun.py`.   
# 
# We will use a cutoff function so that this fine-scale topo matches the linear beach profile of the ocean topography along the edges of this rectangle.  The cutoff is 1 in the center of the rectangle and decays to 0 at the edges:

# In[8]:


# choose DEM grid points:
arcsec13 = 1./(3*3600.)  # 1/3 arcsecond
print('arcsec13 = %.6f degrees = %.2f meters' % (arcsec13,arcsec13*111e3))
x = arange(-100*arcsec13, 150*arcsec13, arcsec13)
y = arange(-55*arcsec13, 55*arcsec13, arcsec13)
X,Y = meshgrid(x,y)
print('X.shape = ', X.shape)

x1,x2 = x.min(), x.max()
y1,y2 = y.min(), y.max()
print('Extent of coastal topo: (%.6f, %.6f, %.6f, %.6f)' % (x1,x2,y1,y2))

# define the cutoff function:

w = 0.001 # width of cutoff layer
cutoff = 1. / (1. + exp(1e4*(X-(x2-w))) + exp(1e4*((x1+w)-X))                   + exp(1e4*(Y-(y2-w))) + exp(1e4*((y1+w)-Y)))


# The topography in this region is the linearly sloping beach augmented by a Gaussian dip.  The beach slope is chosen to agree with the ocean topography offshore (1 km / degree, about 1/100), while onshore there is a smaller slope in this region for illustration.

# In[10]:


Z0 = 1e3*X  # sloping beach matching ocean topography
Z1 = where(X<0, 1e3*X, 0.2e3*X)  # smaller slope on shore
R1 = (X-0.004)**2 + (Y-0.002)**2
Z1 += -4*exp(-500000*R1)         # Gaussian dip
Z = (1-cutoff)*Z0 + cutoff*Z1


# The lower plot in the figure above shows the same topography as on the top, but with x,y units of meters to better show the scale.  Recall that 1 degree is about 111 km and 1/3 arcsec is about 10 meters.
# 
# In the plots above, the red contour is at $Z = 0$, and hence is the "shoreline".  However, the isolated "lake" with elevation $Z < 0$ could be dry land below sea level.  Normally with GeoClaw this region would be filled with water initially up to $Z = 0$ everywhere.  Below in [the Force_Dry section](#force_dry), we discuss how to force this region to be initialized as dry if it is in fact dry land.

# ### Save this as a topofile:

# In[12]:


topo = topotools.Topography()
topo.set_xyZ(x,y,Z)

topofile = '%s/topo_shore.tt3' % inputdir
topo.write(topofile, topo_type=3, Z_format="%11.3e")
print('Created ', topofile)


# In the plot on the left above, the black rectangle showing the extent of the coastal DEM is barely visible.  Zooming in shows that the topography does match up near the edges of this rectangle.  In GeoClaw the finest available topography is used when computing cell-averaged topo values, so the coastal DEM will be used for any cell that overlaps this region. 

# <div id="dtopo"></div>
# 
# ## Create dtopo for an earthquake source:
# 
# We define a simple earthquake in which there is uniform slip on a single subfault. The parameters are chosen to be somewhat reasonable for a subduction zone event offshore, but the shape is a bit odd (width 100 km and length 50 km) in order to give a smallish event with the desired onshore subsidence, for illustration purposes.

# In[14]:


subfault = dtopotools.SubFault()
subfault.strike = 0.
subfault.length = 50.e3
subfault.width = 100.e3
subfault.depth = 10.e3
subfault.slip = 5.
subfault.rake = 90.
subfault.dip = 10.
subfault.longitude = -1.
subfault.latitude = 0.
subfault.coordinate_specification = "top center"

fault = dtopotools.Fault()
fault.subfaults = [subfault]

print("Earthquake magnitude: Mw = %.2f" % fault.Mw())
dtopo_fname = '%s/dtopo_test.tt3' % inputdir
print("Using Okada model to create dtopo file", dtopo_fname)

x_deform = linspace(-2, 1, 100)
y_deform = linspace(-1, 1, 100)
times = [1.]

fault.create_dtopography(x_deform,y_deform,times)
dtopo = fault.dtopo

dtopo.write(dtopo_fname, dtopo_type=3)


# The left plot above shows the sea floor deformation as contours and colors, along with the extent of the continental shelf as blue dashed lines and the shoreline as a red dashed line. The plot on the right shows the vertical deformation along a transect at latitude 0 going through the coastal region of interest.  
# 
# We can compute the subsidence at the location on the shoreline where our fine scale topography is defined as:

# In[16]:


xlon = 0.
ilon = where(dtopo.x<=xlon)[0].max()
ylat = 0.
jlat = where(dtopo.y<=ylat)[0].max()
#print(ilon,jlat)
dz0 = dtopo.dZ[0,jlat,ilon]
print('Surface deformation at x=%.2f, y=%.2f is dz = %.2f meters'       % (xlon,ylat,dz0))


# <div id="force_dry"></div>
# 
# # Force Dry array
# 
# Now suppose that the onshore lake shown in the plots above is really a depression that should be dry land in spite of being below sea level.  We can use the marching front algorithm from [`clawpack.geoclaw.marching_front`](http://depts.washington.edu/clawpack/sampledocs/dev_v5.7.0/marching_front.html) to identify points that are below sea level but disconnected from the coast.  
# 
# We use the marching front algorithm starting by assuming any point with `Z < Z1 = -5` meters should be wet and marching to find all connected points with elevation up to `Z = Z2 = 0`:

# In[18]:


wet_points = marching_front.select_by_flooding(topo.Z, Z1=-5., Z2=0., max_iters=None)


# See the documentation page [Force Cells to be Dry Initially](http://depts.washington.edu/clawpack/sampledocs/dev_v5.7.0/force_dry.html) for more discussion of the cells below...

# ## Create `force_dry_init` array for GeoClaw
# 
# First we buffer the points identified above as discussed in the the documentation page [Force Cells to be Dry Initially](http://depts.washington.edu/clawpack/sampledocs/dev_v5.7.0/force_dry.html).

# In[20]:


dry_points = 1 - wet_points
dry_points_sum = dry_points[1:-1,1:-1] + dry_points[0:-2,1:-1] + dry_points[2:,1:-1] +                  dry_points[1:-1,0:-2] + dry_points[0:-2,0:-2] + dry_points[2:,0:-2] +                  dry_points[1:-1,2:] + dry_points[0:-2,2:] + dry_points[2:,2:]
        
# initialize array to 0 everywhere:
force_dry_init = zeros(dry_points.shape)

# reset in interior to 1 if all points in the 3x3 block around it are dry:
force_dry_init[1:-1,1:-1] = where(dry_points_sum == 9, 1, 0)


# And finally create the input file needed for GeoClaw.  Note that this creates a file with the same format as a topofile having `topo_type == 3` as described in [Topography Data documentation](http://www.clawpack.org/topo.html).  We specify `Z_format= '%1i'` to print out single-digit integers since this file has values 0 or 1 rather than topography elevations (with 1 indicated points that should be forced to be dry when initializing grid patches in GeoClaw).

# In[21]:


force_dry_init_topo = topotools.Topography()
force_dry_init_topo.set_xyZ(topo.x,topo.y,force_dry_init)

fname_force_dry_init = '%s/force_dry_init.tt3' % inputdir
force_dry_init_topo.write(fname_force_dry_init, topo_type=3, Z_format='%1i')
print('Created %s' % fname_force_dry_init)


# See [run_geoclaw.ipynb](run_geoclaw.ipynb) for more discussion and sample GeoClaw results.

# In[ ]:







"""
Module to set up run time parameters for Clawpack -- 1d geoclaw code.

The values set in the function setrun are then written out to data files
that will be read in by the Fortran code.

"""

import os, sys
import numpy as np
from clawpack.geoclaw.nonuniform_grid_tools import make_mapc2p


# Read in nonuniform computational cell edges, which should have
# been created using make_celledges.py:

grid_type = 2
fname_celledges = 'celledges.data'
    
mapc2p, mx_edge, xp_edge = make_mapc2p(fname_celledges)
mx = mx_edge - 1
        
print('Setting mx = %i, cell edges from %g to %g' % (mx,xp_edge[0],xp_edge[-1]))


#------------------------------
def setrun(claw_pkg='geoclaw'):
#------------------------------

    """
    Define the parameters used for running Clawpack.

    INPUT:
        claw_pkg expected to be "geoclaw" for this setrun.

    OUTPUT:
        rundata - object of class ClawRunData

    """

    from clawpack.clawutil import data


    assert claw_pkg.lower() == 'geoclaw',  "Expected claw_pkg = 'geoclaw'"

    num_dim = 1
    rundata = data.ClawRunData(claw_pkg, num_dim)

    #------------------------------------------------------------------
    # Problem-specific parameters to be written to setprob.data:
    #------------------------------------------------------------------
    # Sample setup to write one line to setprob.data ...
    #probdata = rundata.new_UserData(name='probdata',fname='setprob.data')


    #------------------------------------------------------------------
    # Standard Clawpack parameters to be written to claw.data:
    #------------------------------------------------------------------

    clawdata = rundata.clawdata  # initialized when rundata instantiated


    # ---------------
    # Spatial domain:
    # ---------------

    # Number of space dimensions:
    clawdata.num_dim = num_dim

    # Lower and upper edge of computational domain:
    # For nonuniform grid, 0 <= xc <= 1 and the file celledges.data should
    # define the mapping to the physical domain

    clawdata.lower[0] = 0.          # xlower
    clawdata.upper[0] = 1.           # xupper

    # Number of grid cells:
    clawdata.num_cells[0] = mx
    
    # 1D geoclaw requires grid_data:
    rundata.grid_data.grid_type = grid_type  # should be set to 2 above
    rundata.grid_data.fname_celledges = fname_celledges
    rundata.grid_data.monitor_fgmax = False  # record max h,s,etc in each cell?
    rundata.grid_data.monitor_runup = False  # record first and last wet cells?
    rundata.grid_data.monitor_total_zeta = False # record "total mass in wave"?


    # ---------------
    # Size of system:
    # ---------------

    # Number of equations in the system:
    clawdata.num_eqn = 2

    # Number of auxiliary variables in the aux array (initialized in setaux)
    clawdata.num_aux = 2

    # Index of aux array corresponding to capacity function, if there is one:
    clawdata.capa_index = 2


    # -------------
    # Initial time:
    # -------------

    clawdata.t0 = 0.


    # Restart from checkpoint file of a previous run?
    # Note: If restarting, you must also change the Makefile to set:
    #    RESTART = True
    # If restarting, t0 above should be from original run, and the
    # restart_file 'fort.qNNNN' specified below should be in
    # the OUTDIR indicated in Makefile.

    clawdata.restart = False               # True to restart from prior results
    clawdata.restart_file = 'fort.q0006'   # File to use for restart data


    # -------------
    # Output times:
    #--------------

    # Specify at what times the results should be written to fort.q files.
    # Note that the time integration stops after the final output time.

    clawdata.output_style = 1

    if clawdata.output_style==1:
        # Output ntimes frames at equally spaced times up to tfinal:
        # Can specify num_output_times = 0 for no output
        clawdata.num_output_times = 25
        clawdata.tfinal = 750.
        clawdata.output_t0 = True  # output at initial (or restart) time?

    elif clawdata.output_style == 2:
        # Specify a list or numpy array of output times:
        # Include t0 if you want output at the initial time.
        clawdata.output_times =  [0.2,1.6,2.9,4.3,5.6,6.9,8.3,9.6,10.9]

    elif clawdata.output_style == 3:
        # Output every step_interval timesteps over total_steps timesteps:
        clawdata.output_step_interval = 10
        clawdata.total_steps = 20
        clawdata.output_t0 = True  # output at initial (or restart) time?


    clawdata.output_format = 'ascii'  # 'ascii' is only option currently

    clawdata.output_q_components = 'all'   # could be list such as [True,True]
    clawdata.output_aux_components = 'all'  # could be list
    clawdata.output_aux_onlyonce = True   # output aux arrays only at t0


    # ---------------------------------------------------
    # Verbosity of messages to screen during integration:
    # ---------------------------------------------------

    # The current t, dt, and cfl will be printed every time step
    # at AMR levels <= verbosity.  Set verbosity = 0 for no printing.
    #   (E.g. verbosity == 2 means print only on levels 1 and 2.)
    clawdata.verbosity = 0



    # --------------
    # Time stepping:
    # --------------

    # if dt_variable==True:  variable time steps used based on cfl_desired,
    # if dt_variable==False: fixed time steps dt = dt_initial always used.
    clawdata.dt_variable = True

    # Initial time step for variable dt.
    # (If dt_variable==0 then dt=dt_initial for all steps)
    clawdata.dt_initial = 1.

    # Max time step to be allowed if variable dt used:
    clawdata.dt_max = 1.e9

    # Desired Courant number if variable dt used
    clawdata.cfl_desired = 0.90
    # max Courant number to allow without retaking step with a smaller dt:
    clawdata.cfl_max = 1.0

    # Maximum number of time steps to allow between output times:
    clawdata.steps_max = 50000


    # ------------------
    # Method to be used:
    # ------------------

    # Order of accuracy:  1 => Godunov,  2 => Lax-Wendroff plus limiters
    clawdata.order = 2


    # Number of waves in the Riemann solution:
    clawdata.num_waves = 2

    # List of limiters to use for each wave family:
    # Required:  len(limiter) == num_waves
    # Some options:
    #   0 or 'none'     ==> no limiter (Lax-Wendroff)
    #   1 or 'minmod'   ==> minmod
    #   2 or 'superbee' ==> superbee
    #   3 or 'vanleer'  ==> van Leer
    #   4 or 'mc'       ==> MC limiter
    clawdata.limiter = [4,4]

    clawdata.use_fwaves = True    # True ==> use f-wave version of algorithms

    # Source terms splitting:
    #   src_split == 0 or 'none'    ==> no source term (src routine never called)
    #   src_split == 1 or 'godunov' ==> Godunov (1st order) splitting used,
    #   src_split == 2 or 'strang'  ==> Strang (2nd order) splitting used,  not recommended.
    clawdata.source_split = 'godunov'


    # --------------------
    # Boundary conditions:
    # --------------------

    # Number of ghost cells (usually 2)
    clawdata.num_ghost = 2

    # Choice of BCs at xlower and xupper:
    #   0 or 'user'     => user specified (must modify bc1.f to use this option)
    #   1 or 'extrap'   => extrapolation (non-reflecting outflow)
    #   2 or 'periodic' => periodic (must specify this at both boundaries)
    #   3 or 'wall'     => solid wall for systems where q(2) is normal velocity

    clawdata.bc_lower[0] = 'extrap'   # at xlower
    clawdata.bc_upper[0] = 'extrap'   # at xupper


    geo_data = rundata.geo_data

    geo_data.dry_tolerance = 1.e-3

    # Friction source terms:
    #   src_split > 0 required
    #   currently only Manning friction with a single n=friction_coefficient
    #   is supported in 1d.

    geo_data.friction_forcing = True
    geo_data.manning_coefficient =.025

    geo_data.coordinate_system = 1  # linear distance (meters)

    topo_data = rundata.topo_data
    # note that same file can be used for topo as for specifying grid:
    topo_data.topofiles.append([1, fname_celledges])


    # ---------------
    # Gauges:
    # ---------------
    rundata.gaugedata.gauges = []
    # for gauges append lines of the form  [gaugeno, x, t1, t2]

    # for gauges append [gauge id, xc, t1, t2])
    # note that xc is the computational grid point, 0 <= xc <= 1,
    # so if you want to specify physical points xp, these need to be mapped
    # to corresponding xc.


    return rundata


if __name__ == '__main__':
    # Set up run-time parameters and write all data files.
    import sys
    rundata = setrun(*sys.argv[1:])
    rundata.write()




#!/usr/bin/env python
# encoding: utf-8
r"""
Riemann solvers for constant coefficient acoustics.

.. math:: q_t + A q_x = 0
  
where 

.. math:: 
    q(x,t) = \left [ \begin{array}{c} p(x,t) \\ u(x,t) \end{array} \right ]
  
and the coefficient matrix is 

.. math::

    A = \left [\begin{matrix}
    0      & K\\
    1/\rho & 0 
    \end{matrix} \right ]

The parameters :math:`\rho =` density and :math:`K =` bulk modulus are used
to calculate the impedence :math:`= Z` and speed of sound `= c`.

:Authors:
    Kyle T. Mandli (2009-02-03): Initial version
"""
# ============================================================================
#      Copyright (C) 2009 Kyle T. Mandli <mandli@amath.washington.edu>
#
#  Distributed under the terms of the Berkeley Software Distribution (BSD) 
#  license
#                     http://www.opensource.org/licenses/
# ============================================================================

num_eqn = 2
num_waves = 2

def acoustics_1D(q_l,q_r,aux_l,aux_r,problem_data):
    r"""
    Basic 1d acoustics riemann solver, with interleaved arrays
    
    *problem_data* is expected to contain -
     - *zz* - (float) Impedence
     - *cc* - (float) Speed of sound
    
    See :ref:`pyclaw_rp` for more details.
    
    :Version: 1.0 (2009-02-03)
    """
    import numpy as np

    # Convenience
    num_rp = np.size(q_l,1)

    # Return values
    wave = np.empty( (num_eqn, num_waves, num_rp) )
    s = np.empty( (num_waves, num_rp) )
    amdq = np.empty( (num_eqn, num_rp) )
    apdq = np.empty( (num_eqn, num_rp) )
    
    # Local values
    delta = np.empty(np.shape(q_l))
    
    delta = q_r - q_l
    a1 = (-delta[0,:] + problem_data['zz']*delta[1,:]) / (2.0 * problem_data['zz'])
    a2 = (delta[0,:] + problem_data['zz']*delta[1,:]) / (2.0 * problem_data['zz'])
        
    # Compute the waves
    # 1-Wave
    wave[0,0,:] = -a1 * problem_data['zz']
    wave[1,0,:] = a1
    s[0,:] = -problem_data['cc']
        
    # 2-Wave
    wave[0,1,:] = a2 * problem_data['zz']
    wave[1,1,:] = a2
    s[1,:] = problem_data['cc']
    
    # Compute the left going and right going fluctuations
    for m in range(num_eqn):
        amdq[m,:] = s[0,:] * wave[m,0,:]
        apdq[m,:] = s[1,:] * wave[m,1,:]
    
    return wave, s, amdq, apdq



def test_woodward_colella_blast():
    """ Woodward-Colella blast wave test (Euler 1D)"""
    from . import woodward_colella_blast
    from clawpack.pyclaw.util import test_app, check_diff
    import os

    thisdir = os.path.dirname(__file__)

    def verify_woodward_colella_blast(controller):
        """ given an expected value, returns a verification function """
        import numpy as np

        test_solution = controller.solution.state.get_q_global()

        if test_solution is not None:
            expected_density = np.loadtxt(os.path.join(thisdir,'blast_regression_density.txt'))
            test_density = test_solution[0,:]
            return check_diff(expected_density, test_density, reltol=1.e-5,delta=controller.solution.grid.delta)

    return test_app(woodward_colella_blast.setup, verify_woodward_colella_blast, {'outdir': os.path.join(thisdir,'./_output')})





#!/usr/bin/env python
# encoding: utf-8

r"""
One-dimensional acoustics
=========================

Solve the (linear) acoustics equations:

.. math::
    p_t + K u_x & = 0 \\
    u_t + p_x / \rho & = 0.

Here p is the pressure, u is the velocity, K is the bulk modulus,
and :math:`\rho` is the density.

The initial condition is a Gaussian and the boundary conditions are periodic.
The final solution is identical to the initial data because both waves have
crossed the domain exactly once.
"""
from numpy import sqrt, exp, cos
from clawpack import riemann


def setup(use_petsc=False, kernel_language='Fortran', solver_type='classic',
          outdir='./_output', ptwise=False, weno_order=5,
          time_integrator='SSP104', disable_output=False, output_style=1,
          num_cells=100):

    if use_petsc:
        import clawpack.petclaw as pyclaw
    else:
        from clawpack import pyclaw

    if kernel_language == 'Fortran':
        if ptwise:
            riemann_solver = riemann.acoustics_1D_ptwise
        else:
            riemann_solver = riemann.acoustics_1D

    elif kernel_language == 'Python':
        riemann_solver = riemann.acoustics_1D_py.acoustics_1D

    if solver_type == 'classic':
        solver = pyclaw.ClawSolver1D(riemann_solver)
        solver.limiters = pyclaw.limiters.tvd.MC
    elif solver_type == 'sharpclaw':
        solver = pyclaw.SharpClawSolver1D(riemann_solver)
        solver.weno_order = weno_order
        solver.time_integrator = time_integrator
        if time_integrator == 'SSPLMMk3':
            solver.lmm_steps = 4
    else:
        raise Exception('Unrecognized value of solver_type.')

    solver.kernel_language = kernel_language

    x = pyclaw.Dimension(0.0, 1.0, num_cells, name='x')
    domain = pyclaw.Domain(x)
    num_eqn = 2
    state = pyclaw.State(domain, num_eqn)

    solver.bc_lower[0] = pyclaw.BC.periodic
    solver.bc_upper[0] = pyclaw.BC.periodic

    rho = 1.0   # Material density
    bulk = 1.0  # Material bulk modulus

    state.problem_data['rho'] = rho
    state.problem_data['bulk'] = bulk
    state.problem_data['zz'] = sqrt(rho*bulk)   # Impedance
    state.problem_data['cc'] = sqrt(bulk/rho)   # Sound speed

    xc = domain.grid.x.centers
    beta = 200
    gamma = 0
    x0 = 0.5
    state.q[0, :] = exp(-beta * (xc-x0)**2) * cos(gamma * (xc - x0))
    state.q[1, :] = 0.0

    solver.dt_initial = domain.grid.delta[0] / state.problem_data['cc'] * 0.1

    claw = pyclaw.Controller()
    claw.solution = pyclaw.Solution(state, domain)
    claw.solver = solver
    claw.outdir = outdir
    claw.output_style = output_style
    if output_style == 1:
        claw.tfinal = 1.0
        claw.num_output_times = 10
    elif output_style == 3:
        claw.nstep = 1
        claw.num_output_times = 1
    claw.keep_copy = True
    if disable_output:
        claw.output_format = None
    claw.setplot = setplot

    return claw


def setplot(plotdata):
    """
    Specify what is to be plotted at each frame.
    Input:  plotdata, an instance of visclaw.data.ClawPlotData.
    Output: a modified version of plotdata.
    """
    plotdata.clearfigures()  # clear any old figures,axes,items data

    # Figure for pressure
    plotfigure = plotdata.new_plotfigure(name='Pressure', figno=1)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.axescmd = 'subplot(211)'
    plotaxes.ylimits = [-0.2, 1.0]
    plotaxes.title = 'Pressure'

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = 0
    plotitem.plotstyle = '-o'
    plotitem.color = 'b'
    plotitem.kwargs = {'linewidth': 2, 'markersize': 5}

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.axescmd = 'subplot(212)'
    plotaxes.xlimits = 'auto'
    plotaxes.ylimits = [-0.5, 1.1]
    plotaxes.title = 'Velocity'

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = 1
    plotitem.plotstyle = '-'
    plotitem.color = 'b'
    plotitem.kwargs = {'linewidth': 3, 'markersize': 5}

    return plotdata


def run_and_plot(**kwargs):
    claw = setup(kwargs)
    claw.run()
    from clawpack.pyclaw import plot
    plot.interactive_plot(setplot=setplot)

if __name__ == "__main__":
    from clawpack.pyclaw.util import run_app_from_main
    output = run_app_from_main(setup, setplot)




""" 
Set up the plot figures, axes, and items to be done for each frame.

This module is imported by the plotting routines and then the
function setplot is called to set the plot parameters.
    
""" 

import numpy as np
import matplotlib.pyplot as plt

from clawpack.geoclaw import topotools

try:
    TG32412 = np.loadtxt('32412_notide.txt')
except:
    print("*** Could not load DART data file")

#--------------------------
def setplot(plotdata=None):
#--------------------------
    
    """ 
    Specify what is to be plotted at each frame.
    Input:  plotdata, an instance of pyclaw.plotters.data.ClawPlotData.
    Output: a modified version of plotdata.
    
    """ 


    from clawpack.visclaw import colormaps, geoplot
    from numpy import linspace

    if plotdata is None:
        from clawpack.visclaw.data import ClawPlotData
        plotdata = ClawPlotData()


    plotdata.clearfigures()  # clear any old figures,axes,items data
    plotdata.format = 'ascii'    # 'ascii' or 'binary' to match setrun.py


    # To plot gauge locations on pcolor or contour plot, use this as
    # an afteraxis function:

    def addgauges(current_data):
        from clawpack.visclaw import gaugetools
        gaugetools.plot_gauge_locations(current_data.plotdata, \
             gaugenos='all', format_string='ko', add_labels=True)
    

    #-----------------------------------------
    # Figure for surface
    #-----------------------------------------
    plotfigure = plotdata.new_plotfigure(name='Surface', figno=0)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes('pcolor')
    plotaxes.title = 'Surface at time h:m:s'
    plotaxes.aspect_latitude = -30.
    plotaxes.xticks_fontsize = 10
    plotaxes.yticks_fontsize = 10
    plotaxes.xlabel = 'longitude'
    plotaxes.ylabel = 'latitude'

    plotaxes.afteraxes = addgauges

    # Water
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    #plotitem.plot_var = geoplot.surface
    plotitem.plot_var = geoplot.surface_or_depth
    plotitem.pcolor_cmap = geoplot.tsunami_colormap
    plotitem.pcolor_cmin = -0.2
    plotitem.pcolor_cmax = 0.2
    plotitem.add_colorbar = True
    plotitem.colorbar_shrink = 0.8
    plotitem.colorbar_extend = 'both'
    plotitem.amr_celledges_show = [0,0,0]
    plotitem.patchedges_show = 1

    # Land
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.plot_var = geoplot.land
    plotitem.pcolor_cmap = geoplot.land_colors
    plotitem.pcolor_cmin = 0.0
    plotitem.pcolor_cmax = 100.0
    plotitem.add_colorbar = False
    plotitem.amr_celledges_show = [1,1,0]
    plotitem.patchedges_show = 1
    plotaxes.xlimits = [-120,-60]
    plotaxes.ylimits = [-60,0]

    # add contour lines of bathy if desired:
    plotitem = plotaxes.new_plotitem(plot_type='2d_contour')
    plotitem.show = False
    plotitem.plot_var = geoplot.topo
    plotitem.contour_levels = linspace(-3000,-3000,1)
    plotitem.amr_contour_colors = ['y']  # color on each level
    plotitem.kwargs = {'linestyles':'solid','linewidths':2}
    plotitem.amr_contour_show = [1,0,0]  
    plotitem.celledges_show = 0
    plotitem.patchedges_show = 0


    #-----------------------------------------
    # Figures for gauges
    #-----------------------------------------
    plotfigure = plotdata.new_plotfigure(name='Surface at gauges', figno=300, \
                    type='each_gauge')
    plotfigure.clf_each_gauge = True

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.time_scale = 1/3600.  # convert to hours
    plotaxes.xlimits = [0, 9]
    plotaxes.ylimits = [-0.3, 0.3]
    plotaxes.title = 'Surface'
    plotaxes.title_fontsize = 15
    plotaxes.time_label = 'time (hours)'
    plotaxes.ylabel = 'surface elevation (m)'
    plotaxes.grid = True

    def add_obs(current_data):
        from pylab import plot, legend
        gaugeno = current_data.gaugeno
        if gaugeno == 32412:
            try:
                tgauge = TG32412[:,0] / 3600.  # convert to hours
                plot(tgauge, TG32412[:,1], 'r', linewidth=1.0)
                legend(['GeoClaw','Obs'],loc='lower right')
            except: pass

    plotaxes.afteraxes = add_obs

    # Plot surface as blue curve:
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = 3  # eta
    plotitem.plotstyle = 'b-'
    plotitem.kwargs = {'linewidth':1.8}


    # Plot topo as green curve:
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.show = False  # not being used

    def gaugetopo(current_data):
        q = current_data.q
        h = q[0,:]
        eta = q[3,:]
        topo = eta - h
        return topo
        
    plotitem.plot_var = gaugetopo
    plotitem.plotstyle = 'g-'



    #-----------------------------------------
    # Plots of timing (CPU and wall time):

    def make_timing_plots(plotdata):
        from clawpack.visclaw import plot_timing_stats
        import os,sys
        try:
            timing_plotdir = plotdata.plotdir + '/_timing_figures'
            os.system('mkdir -p %s' % timing_plotdir)
            # adjust units for plots based on problem:
            units = {'comptime':'seconds', 'simtime':'hours', 
                     'cell':'millions'}
            plot_timing_stats.make_plots(outdir=plotdata.outdir, 
                                          make_pngs=True,
                                          plotdir=timing_plotdir, 
                                          units=units)
        except:
            print('*** Error making timing plots')

    otherfigure = plotdata.new_otherfigure(name='timing plots',
                    fname='_timing_figures/timing.html')
    otherfigure.makefig = make_timing_plots


    #-----------------------------------------
    
    # Parameters used only when creating html and/or latex hardcopy
    # e.g., via pyclaw.plotters.frametools.printframes:

    plotdata.printfigs = True                # print figures
    plotdata.print_format = 'png'            # file format
    plotdata.print_framenos = 'all'          # list of frames to print
    plotdata.print_gaugenos = 'all'          # list of gauges to print
    plotdata.print_fignos = 'all'            # list of figures to print
    plotdata.html = True                     # create html files of plots?
    plotdata.html_homelink = '../README.html'   # pointer for top of index
    plotdata.latex = True                    # create latex file of plots?
    plotdata.latex_figsperline = 2           # layout of plots
    plotdata.latex_framesperline = 1         # layout of plots
    plotdata.latex_makepdf = False           # also run pdflatex?
    plotdata.parallel = True                 # make multiple frame png's at once

    return plotdata




from __future__ import absolute_import
from __future__ import print_function
import RiemannSolver
from RiemannSolver import *
from numpy import load, absolute


tolerance = 0.00005
print("This script asserts the results for input read from data files q.npy and aux.npy with tolerance =", tolerance)



rs = RiemannSolver( 10, 2, 2**15, 2, 2)
    
rs.q = load("q.npy")
rs.aux = load("aux.npy")

wavesRead = load("waves.npy")
sRead = load("s.npy")

waves1, s1  = rs.solveVectorized(timer = True)
#waves2, s2  = rs.solvePointwize(timer = True)

assert (absolute(wavesRead - waves1)< tolerance).all()
assert (absolute(sRead -s1)< tolerance).all()

